from __future__ import annotations

import os
import typing as t
from pathlib import Path

import duckdb
import pytest
from pyspark import SparkConf
from pyspark.sql import SparkSession as PySparkSession
from pytest_postgresql.janitor import DatabaseJanitor

from sqlframe.base.session import _BaseSession
from sqlframe.bigquery.session import BigQuerySession
from sqlframe.databricks.session import DatabricksSession
from sqlframe.duckdb.session import DuckDBSession
from sqlframe.postgres.session import PostgresSession
from sqlframe.redshift.session import RedshiftSession
from sqlframe.snowflake.session import SnowflakeSession
from sqlframe.spark.session import SparkSession
from sqlframe.standalone import types as StandaloneTypes
from sqlframe.standalone.dataframe import StandaloneDataFrame
from sqlframe.standalone.session import StandaloneSession

if t.TYPE_CHECKING:
    from databricks.sql import Connection as DatabricksConnection
    from google.cloud.bigquery.dbapi.connection import (
        Connection as BigQueryConnection,
    )
    from redshift_connector.core import Connection as RedshiftConnection
    from snowflake.connector import SnowflakeConnection

    from tests.types import EmployeeData


def load_tpcds(paths: t.List[Path], session: t.Union[_BaseSession, PySparkSession]):
    for path in paths:
        table_name = path.name
        session.read.parquet(str(path / "*.parquet")).createOrReplaceTempView(table_name)


@pytest.fixture(scope="session")
def gen_tpcds(tmp_path_factory) -> t.List[Path]:
    path_root = tmp_path_factory.mktemp("tpcds")
    results = []
    tables = [
        "web_site",
        "web_sales",
        "web_returns",
        "web_page",
        "warehouse",
        "time_dim",
        "store_sales",
        "store_returns",
        "store",
        "ship_mode",
        "reason",
        "promotion",
        "item",
        "inventory",
        "income_band",
        "household_demographics",
        "date_dim",
        "customer_demographics",
        "customer_address",
        "customer",
        "catalog_sales",
        "catalog_returns",
        "catalog_page",
        "call_center",
    ]
    con = duckdb.connect()
    con.sql("CALL dsdgen(sf=0.01)")
    for table in tables:
        path = path_root / table
        path.mkdir(parents=True)
        con.sql(
            f"COPY (SELECT * FROM {table}) TO '{path}' (FORMAT PARQUET, PER_THREAD_OUTPUT TRUE)"
        )
        results.append(path)
    con.close()
    return results


@pytest.fixture(scope="session")
def pyspark_session(tmp_path_factory, gen_tpcds: t.List[Path]) -> PySparkSession:
    data_dir = tmp_path_factory.mktemp("spark_connection")
    derby_dir = tmp_path_factory.mktemp("derby")
    spark = (
        PySparkSession.builder.enableHiveSupport()
        .config("spark.sql.analyzer.failAmbiguousSelfJoin", "false")
        .config("spark.sql.warehouse.dir", data_dir)
        .config("spark.driver.extraJavaOptions", f"-Dderby.system.home={derby_dir}")
        .config("spark.sql.shuffle.partitions", 1)
        .config("spark.sql.session.timeZone", "UTC")
        .master("local[1]")
        .appName("Unit-tests")
        .getOrCreate()
    )
    spark.sparkContext.setLogLevel("ERROR")
    spark.sql("DROP DATABASE IF EXISTS db1 CASCADE")
    spark.sql("CREATE DATABASE db1")
    spark.sql("CREATE TABLE db1.table1 (id INTEGER, name STRING)")
    spark.catalog.registerFunction("add", lambda x, y: x + y)
    spark.catalog.setCurrentDatabase("db1")
    load_tpcds(gen_tpcds, spark)
    return spark


@pytest.fixture(scope="function")
def standalone_session() -> StandaloneSession:
    return StandaloneSession()


@pytest.fixture()
def spark_session(pyspark_session: PySparkSession) -> SparkSession:
    return SparkSession()


@pytest.fixture(scope="function")
def duckdb_session() -> DuckDBSession:
    connector = duckdb.connect()
    connector.execute("set TimeZone = 'UTC'")
    connector.execute("SELECT * FROM duckdb_settings() WHERE name = 'TimeZone'")
    assert connector.fetchone()[1] == "UTC"  # type: ignore
    return DuckDBSession(conn=connector)


@pytest.fixture
def function_scoped_postgres(postgresql_proc):
    import psycopg2

    janitor = DatabaseJanitor(
        user=postgresql_proc.user,
        host=postgresql_proc.host,
        port=postgresql_proc.port,
        dbname=postgresql_proc.dbname,
        version=postgresql_proc.version,
        password=postgresql_proc.password,
    )
    with janitor:
        conn = psycopg2.connect(
            dbname=postgresql_proc.dbname,
            user=postgresql_proc.user,
            password=postgresql_proc.password,
            host=postgresql_proc.host,
            port=postgresql_proc.port,
        )
        yield conn


@pytest.fixture
def postgres_session(function_scoped_postgres) -> PostgresSession:
    session = PostgresSession(function_scoped_postgres)
    session._execute("SET TIME ZONE 'UTC'")
    session._execute('CREATE SCHEMA "db1"')
    session._execute('CREATE TABLE "table1" (id INTEGER, name VARCHAR(100))')
    session._execute(
        "CREATE FUNCTION testing(integer, integer) RETURNS integer AS 'select $1 + $2;' LANGUAGE SQL"
    )
    return session


@pytest.fixture(scope="session")
def bigquery_connection() -> BigQueryConnection:
    from google.cloud.bigquery.dbapi import connect

    connection = connect()
    cursor = connection.cursor()
    cursor.execute("DROP SCHEMA IF EXISTS sqlframe.db1 CASCADE")
    cursor.execute("CREATE SCHEMA IF NOT EXISTS sqlframe.db1")
    cursor.execute("CREATE TABLE IF NOT EXISTS sqlframe.db1.table1 (id INTEGER, name STRING)")
    cursor.execute(
        "CREATE FUNCTION IF NOT EXISTS sqlframe.db1.add(x INT64, y INT64) RETURNS INT64 AS (x + y);"
    )
    return connection


@pytest.fixture
def bigquery_session(bigquery_connection: BigQueryConnection) -> BigQuerySession:
    return BigQuerySession(bigquery_connection, default_dataset="sqlframe.db1")


@pytest.fixture(scope="session")
def redshift_connection() -> RedshiftConnection:
    from redshift_connector import connect

    conn = connect(
        user=os.environ["SQLFRAME_REDSHIFT_USER"],
        password=os.environ["SQLFRAME_REDSHIFT_PASSWORD"],
        database=os.environ["SQLFRAME_REDSHIFT_DATABASE"],
        host=os.environ["SQLFRAME_REDSHIFT_HOST"],
        port=int(os.environ["SQLFRAME_REDSHIFT_PORT"]),
    )
    conn.autocommit = True
    return conn


@pytest.fixture
def redshift_session(redshift_connection: RedshiftConnection) -> RedshiftSession:
    session = RedshiftSession(redshift_connection)
    session._execute("CREATE SCHEMA IF NOT EXISTS db1")
    session._execute("CREATE TABLE IF NOT EXISTS db1.table1 (id INTEGER, name VARCHAR(100))")
    session._execute(
        "CREATE OR REPLACE FUNCTION db1.add(x INT, y INT) RETURNS INT IMMUTABLE AS $$ select $1 + $2 $$ LANGUAGE SQL"
    )
    return session


@pytest.fixture(scope="session")
def snowflake_connection() -> SnowflakeConnection:
    from snowflake.connector import connect

    conn = connect(
        account=os.environ["SQLFRAME_SNOWFLAKE_ACCOUNT"],
        user=os.environ["SQLFRAME_SNOWFLAKE_USER"],
        password=os.environ["SQLFRAME_SNOWFLAKE_PASSWORD"],
        warehouse=os.environ["SQLFRAME_SNOWFLAKE_WAREHOUSE"],
        database=os.environ["SQLFRAME_SNOWFLAKE_DATABASE"],
        schema=os.environ["SQLFRAME_SNOWFLAKE_SCHEMA"],
    )
    return conn


@pytest.fixture
def snowflake_session(snowflake_connection: SnowflakeConnection) -> SnowflakeSession:
    session = SnowflakeSession(snowflake_connection)
    session._execute("ALTER SESSION SET TIMEZONE = 'UTC'")
    session._execute("CREATE SCHEMA IF NOT EXISTS db1")
    session._execute("CREATE TABLE IF NOT EXISTS db1.table1 (id INTEGER, name VARCHAR(100))")
    session._execute(
        "CREATE OR REPLACE FUNCTION db1.add(x INT, y INT) RETURNS INT IMMUTABLE AS $$ select x + y $$"
    )
    return session


@pytest.fixture(scope="session")
def databricks_connection() -> DatabricksConnection:
    from databricks.sql import connect

    conn = connect(
        server_hostname=os.environ["SQLFRAME_DATABRICKS_SERVER_HOSTNAME"],
        http_path=os.environ["SQLFRAME_DATABRICKS_HTTP_PATH"],
        access_token=os.environ["SQLFRAME_DATABRICKS_ACCESS_TOKEN"],
        auth_type="access_token",
        catalog=os.environ["SQLFRAME_DATABRICKS_CATALOG"],
        schema=os.environ["SQLFRAME_DATABRICKS_SCHEMA"],
        _disable_pandas=True,
    )
    return conn


@pytest.fixture
def databricks_session(databricks_connection: DatabricksConnection) -> DatabricksSession:
    session = DatabricksSession(databricks_connection)
    session._execute("CREATE SCHEMA IF NOT EXISTS db1")
    session._execute("CREATE TABLE IF NOT EXISTS db1.table1 (id INTEGER, name VARCHAR(100))")
    session._execute("CREATE OR REPLACE FUNCTION db1.add(x INT, y INT) RETURNS INT RETURN x + y")
    return session


@pytest.fixture(scope="module")
def _employee_data() -> EmployeeData:
    return [
        (1, "Jack", "Shephard", 37, 1),
        (2, "John", "Locke", 65, 1),
        (3, "Kate", "Austen", 37, 2),
        (4, "Claire", "Littleton", 27, 2),
        (5, "Hugo", "Reyes", 29, 100),
    ]


@pytest.fixture(scope="function")
def standalone_employee(
    standalone_session: StandaloneSession, _employee_data: EmployeeData
) -> StandaloneDataFrame:
    sqlf_employee_schema = StandaloneTypes.StructType(
        [
            StandaloneTypes.StructField("employee_id", StandaloneTypes.IntegerType(), False),
            StandaloneTypes.StructField("fname", StandaloneTypes.StringType(), False),
            StandaloneTypes.StructField("lname", StandaloneTypes.StringType(), False),
            StandaloneTypes.StructField("age", StandaloneTypes.IntegerType(), False),
            StandaloneTypes.StructField("store_id", StandaloneTypes.IntegerType(), False),
        ]
    )
    df = standalone_session.createDataFrame(data=_employee_data, schema=sqlf_employee_schema)
    standalone_session.catalog.add_table("employee", sqlf_employee_schema)  # type: ignore
    return df



from __future__ import annotations

import typing as t
from functools import cached_property

from sqlframe.base.session import _BaseSession
from sqlframe.base.util import soundex, verify_numpy_installed
from sqlframe.duckdb.catalog import DuckDBCatalog
from sqlframe.duckdb.dataframe import DuckDBDataFrame
from sqlframe.duckdb.readwriter import (
    DuckDBDataFrameReader,
    DuckDBDataFrameWriter,
)
from sqlframe.duckdb.udf import DuckDBUDFRegistration

if t.TYPE_CHECKING:
    from duckdb import DuckDBPyConnection

else:
    DuckDBPyConnection = t.Any


class DuckDBSession(
    _BaseSession[  # type: ignore
        DuckDBCatalog,
        DuckDBDataFrameReader,
        DuckDBDataFrameWriter,
        DuckDBDataFrame,
        DuckDBPyConnection,
        DuckDBUDFRegistration,
    ]
):
    _catalog = DuckDBCatalog
    _reader = DuckDBDataFrameReader
    _writer = DuckDBDataFrameWriter
    _df = DuckDBDataFrame
    _udf_registration = DuckDBUDFRegistration

    def __init__(self, conn: t.Optional[DuckDBPyConnection] = None, *args, **kwargs):
        import duckdb
        from duckdb.typing import VARCHAR

        if not hasattr(self, "_conn"):
            conn = conn or duckdb.connect()
            try:
                # Creating a function requires numpy to be installed so if they don't have it, we'll just skip it
                verify_numpy_installed()
                conn.create_function("SOUNDEX", lambda x: soundex(x), return_type=VARCHAR)
            except ImportError:
                pass

            super().__init__(conn, *args, **kwargs)
            self._last_result = None

    @cached_property
    def _cur(self) -> DuckDBPyConnection:  # type: ignore
        return self._conn

    @classmethod
    def _try_get_map(cls, value: t.Any) -> t.Optional[t.Dict[str, t.Any]]:
        if value and isinstance(value, dict):
            # DuckDB < 1.1.0 support
            if "key" in value and "value" in value:
                return dict(zip(value["key"], value["value"]))
            # DuckDB >= 1.1.0 support
            # If a key is not a string then it must not represent a column and therefore must be a map
            if len([k for k in value if not isinstance(k, str)]) > 0:
                return value
        return None

    def _execute(self, sql: str) -> None:
        self._last_result = self._cur.execute(sql)  # type: ignore

    class Builder(_BaseSession.Builder):
        DEFAULT_EXECUTION_DIALECT = "duckdb"

        @cached_property
        def session(self) -> DuckDBSession:
            return DuckDBSession(**self._session_kwargs)

        def getOrCreate(self) -> DuckDBSession:
            return super().getOrCreate()  # type: ignore

    builder = Builder()



import pytest

from sqlframe.spark import Column as SparkColumn
from sqlframe.spark import Row as SparkRow
from sqlframe.spark import (
    SparkCatalog,
    SparkDataFrame,
    SparkDataFrameNaFunctions,
    SparkDataFrameReader,
    SparkDataFrameStatFunctions,
    SparkDataFrameWriter,
    SparkGroupedData,
    SparkSession,
    SparkUDFRegistration,
)
from sqlframe.spark import Window as SparkWindow
from sqlframe.spark import WindowSpec as SparkWindowSpec
from sqlframe.spark import functions as SparkF
from sqlframe.spark import types as SparkTypes


@pytest.mark.forked
def test_replace_pyspark_spark(check_pyspark_imports):
    check_pyspark_imports(
        "spark",
        sqlf_session=SparkSession,
        sqlf_catalog=SparkCatalog,
        sqlf_column=SparkColumn,
        sqlf_dataframe=SparkDataFrame,
        sqlf_grouped_data=SparkGroupedData,
        sqlf_window=SparkWindow,
        sqlf_window_spec=SparkWindowSpec,
        sqlf_functions=SparkF,
        sqlf_types=SparkTypes,
        sqlf_udf_registration=SparkUDFRegistration,
        sqlf_dataframe_reader=SparkDataFrameReader,
        sqlf_dataframe_writer=SparkDataFrameWriter,
        sqlf_dataframe_na_functions=SparkDataFrameNaFunctions,
        sqlf_dataframe_stat_functions=SparkDataFrameStatFunctions,
        sqlf_row=SparkRow,
    )



from __future__ import annotations

import typing as t

from sqlframe.base.types import Row

if t.TYPE_CHECKING:
    from sqlframe.base.session import _BaseSession

pytest_plugins = ["tests.integration.fixtures"]


def test_load_no_format(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df = session.read.load("tests/fixtures/employee.json")
    assert df.collect() == [
        Row(**{"employee_id": 1, "fname": "Jack", "lname": "Shephard", "age": 37, "store_id": 1}),
        Row(**{"employee_id": 2, "fname": "John", "lname": "Locke", "age": 65, "store_id": 1}),
        Row(**{"employee_id": 3, "fname": "Kate", "lname": "Austen", "age": 37, "store_id": 2}),
        Row(
            **{"employee_id": 4, "fname": "Claire", "lname": "Littleton", "age": 27, "store_id": 2}
        ),
        Row(**{"employee_id": 5, "fname": "Hugo", "lname": "Reyes", "age": 29, "store_id": 100}),
    ]


def test_load_no_format_schema(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df = session.read.load(
        "tests/fixtures/employee.json",
        schema="employee_id STRING, fname STRING, lname STRING, age INT, store_id INT",
    )
    assert df.collect() == [
        Row(**{"employee_id": "1", "fname": "Jack", "lname": "Shephard", "age": 37, "store_id": 1}),
        Row(**{"employee_id": "2", "fname": "John", "lname": "Locke", "age": 65, "store_id": 1}),
        Row(**{"employee_id": "3", "fname": "Kate", "lname": "Austen", "age": 37, "store_id": 2}),
        Row(
            **{
                "employee_id": "4",
                "fname": "Claire",
                "lname": "Littleton",
                "age": 27,
                "store_id": 2,
            }
        ),
        Row(**{"employee_id": "5", "fname": "Hugo", "lname": "Reyes", "age": 29, "store_id": 100}),
    ]


def test_load_json(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df = session.read.load("tests/fixtures/employee.json", format="json")
    assert df.collect() == [
        Row(**{"employee_id": 1, "fname": "Jack", "lname": "Shephard", "age": 37, "store_id": 1}),
        Row(**{"employee_id": 2, "fname": "John", "lname": "Locke", "age": 65, "store_id": 1}),
        Row(**{"employee_id": 3, "fname": "Kate", "lname": "Austen", "age": 37, "store_id": 2}),
        Row(
            **{"employee_id": 4, "fname": "Claire", "lname": "Littleton", "age": 27, "store_id": 2}
        ),
        Row(**{"employee_id": 5, "fname": "Hugo", "lname": "Reyes", "age": 29, "store_id": 100}),
    ]


def test_json(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df = session.read.json("tests/fixtures/employee.json")
    assert df.collect() == [
        Row(**{"employee_id": 1, "fname": "Jack", "lname": "Shephard", "age": 37, "store_id": 1}),
        Row(**{"employee_id": 2, "fname": "John", "lname": "Locke", "age": 65, "store_id": 1}),
        Row(**{"employee_id": 3, "fname": "Kate", "lname": "Austen", "age": 37, "store_id": 2}),
        Row(
            **{"employee_id": 4, "fname": "Claire", "lname": "Littleton", "age": 27, "store_id": 2}
        ),
        Row(**{"employee_id": 5, "fname": "Hugo", "lname": "Reyes", "age": 29, "store_id": 100}),
    ]


def test_load_parquet(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df = session.read.load("tests/fixtures/employee.parquet", format="parquet")
    assert df.collect() == [
        Row(**{"employee_id": 1, "fname": "Jack", "lname": "Shephard", "age": 37, "store_id": 1}),
        Row(**{"employee_id": 2, "fname": "John", "lname": "Locke", "age": 65, "store_id": 1}),
        Row(**{"employee_id": 3, "fname": "Kate", "lname": "Austen", "age": 37, "store_id": 2}),
        Row(
            **{"employee_id": 4, "fname": "Claire", "lname": "Littleton", "age": 27, "store_id": 2}
        ),
        Row(**{"employee_id": 5, "fname": "Hugo", "lname": "Reyes", "age": 29, "store_id": 100}),
    ]


def test_parquet(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df = session.read.parquet("tests/fixtures/employee.parquet")
    assert df.collect() == [
        Row(**{"employee_id": 1, "fname": "Jack", "lname": "Shephard", "age": 37, "store_id": 1}),
        Row(**{"employee_id": 2, "fname": "John", "lname": "Locke", "age": 65, "store_id": 1}),
        Row(**{"employee_id": 3, "fname": "Kate", "lname": "Austen", "age": 37, "store_id": 2}),
        Row(
            **{"employee_id": 4, "fname": "Claire", "lname": "Littleton", "age": 27, "store_id": 2}
        ),
        Row(**{"employee_id": 5, "fname": "Hugo", "lname": "Reyes", "age": 29, "store_id": 100}),
    ]


def test_load_csv(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df = session.read.load("tests/fixtures/employee.csv", format="csv")
    assert df.collect() == [
        Row(**{"employee_id": 1, "fname": "Jack", "lname": "Shephard", "age": 37, "store_id": 1}),
        Row(**{"employee_id": 2, "fname": "John", "lname": "Locke", "age": 65, "store_id": 1}),
        Row(**{"employee_id": 3, "fname": "Kate", "lname": "Austen", "age": 37, "store_id": 2}),
        Row(
            **{"employee_id": 4, "fname": "Claire", "lname": "Littleton", "age": 27, "store_id": 2}
        ),
        Row(**{"employee_id": 5, "fname": "Hugo", "lname": "Reyes", "age": 29, "store_id": 100}),
    ]


def test_csv(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df = session.read.csv("tests/fixtures/employee.csv")
    assert df.collect() == [
        Row(**{"employee_id": 1, "fname": "Jack", "lname": "Shephard", "age": 37, "store_id": 1}),
        Row(**{"employee_id": 2, "fname": "John", "lname": "Locke", "age": 65, "store_id": 1}),
        Row(**{"employee_id": 3, "fname": "Kate", "lname": "Austen", "age": 37, "store_id": 2}),
        Row(
            **{"employee_id": 4, "fname": "Claire", "lname": "Littleton", "age": 27, "store_id": 2}
        ),
        Row(**{"employee_id": 5, "fname": "Hugo", "lname": "Reyes", "age": 29, "store_id": 100}),
    ]



import findspark
import pytest

from sqlframe import activate, deactivate


@pytest.fixture()
def check_pyspark_imports():
    def _check_pyspark_imports(
        engine_name,
        sqlf_session,
        sqlf_catalog,
        sqlf_column,
        sqlf_dataframe,
        sqlf_grouped_data,
        sqlf_window,
        sqlf_window_spec,
        sqlf_functions,
        sqlf_types,
        sqlf_udf_registration,
        sqlf_dataframe_reader,
        sqlf_dataframe_writer,
        sqlf_dataframe_na_functions,
        sqlf_dataframe_stat_functions,
        sqlf_row,
    ):
        activate(engine=engine_name)
        findspark.init()
        from pyspark.sql import (
            Catalog,
            Column,
            DataFrame,
            DataFrameNaFunctions,
            DataFrameReader,
            DataFrameStatFunctions,
            DataFrameWriter,
            GroupedData,
            Row,
            SparkSession,
            UDFRegistration,
            Window,
            WindowSpec,
            types,
        )
        from pyspark.sql import functions as F

        assert SparkSession == sqlf_session
        assert Catalog == sqlf_catalog
        assert Column == sqlf_column
        assert GroupedData == sqlf_grouped_data
        assert DataFrame == sqlf_dataframe
        assert Window == sqlf_window
        assert WindowSpec == sqlf_window_spec
        assert F == sqlf_functions
        assert types == sqlf_types
        assert UDFRegistration == sqlf_udf_registration
        assert DataFrameNaFunctions == sqlf_dataframe_na_functions
        assert DataFrameStatFunctions == sqlf_dataframe_stat_functions
        assert DataFrameReader == sqlf_dataframe_reader
        assert DataFrameWriter == sqlf_dataframe_writer
        assert Row == sqlf_row

        from pyspark.sql.session import SparkSession as SparkSession2

        assert SparkSession2 == sqlf_session

        from pyspark.sql.catalog import Catalog as Catalog2

        assert Catalog2 == sqlf_catalog

        from pyspark.sql.column import Column as Column2

        assert Column2 == sqlf_column

        from pyspark.sql.dataframe import DataFrame as DataFrame2
        from pyspark.sql.dataframe import DataFrameNaFunctions as DataFrameNaFunctions2
        from pyspark.sql.dataframe import (
            DataFrameStatFunctions as DataFrameStatFunctions2,
        )

        assert DataFrame2 == sqlf_dataframe
        assert DataFrameNaFunctions2 == sqlf_dataframe_na_functions
        assert DataFrameStatFunctions2 == sqlf_dataframe_stat_functions

        from pyspark.sql.group import GroupedData as GroupedData2

        assert GroupedData2 == sqlf_grouped_data

        from pyspark.sql.window import WindowSpec as WindowSpec2

        assert WindowSpec2 == sqlf_window_spec

        from pyspark.sql.readwriter import DataFrameReader as DataFrameReader2
        from pyspark.sql.readwriter import DataFrameWriter as DataFrameWriter2

        assert DataFrameReader2 == sqlf_dataframe_reader
        assert DataFrameWriter2 == sqlf_dataframe_writer

        from pyspark.sql.window import Window, WindowSpec

        assert Window == sqlf_window
        assert WindowSpec == sqlf_window_spec

        from pyspark.sql import functions as F
        from pyspark.sql import types

        assert F == sqlf_functions
        assert types == sqlf_types
        assert types.Row == sqlf_row

        from pyspark.sql import UDFRegistration

        assert UDFRegistration == sqlf_udf_registration

        import pyspark.sql.functions as F
        import pyspark.sql.types as types

        assert F == sqlf_functions
        assert types == sqlf_types

        deactivate()
        findspark.init()
        from pyspark.sql import (
            Catalog,
            Column,
            DataFrame,
            DataFrameNaFunctions,
            DataFrameReader,
            DataFrameStatFunctions,
            DataFrameWriter,
            GroupedData,
            Row,
            SparkSession,
            UDFRegistration,
            Window,
            WindowSpec,
            types,
        )
        from pyspark.sql import functions as F

        assert SparkSession != sqlf_session
        assert Catalog != sqlf_catalog
        assert Column != sqlf_column
        assert GroupedData != sqlf_grouped_data
        assert DataFrame != sqlf_dataframe
        assert Window != sqlf_window
        assert WindowSpec != sqlf_window_spec
        assert F != sqlf_functions
        assert types != sqlf_types
        assert UDFRegistration != sqlf_udf_registration
        assert DataFrameNaFunctions != sqlf_dataframe_na_functions
        assert DataFrameStatFunctions != sqlf_dataframe_stat_functions
        assert DataFrameReader != sqlf_dataframe_reader
        assert DataFrameWriter != sqlf_dataframe_writer
        assert Row != sqlf_row

        assert F.JVMView is not None
        assert SparkSession.readStream is not None

        from pyspark.sql.session import SparkSession as SparkSession2

        assert SparkSession2 != sqlf_session

        from pyspark.sql.catalog import Catalog as Catalog2

        assert Catalog2 != sqlf_catalog

        from pyspark.sql.column import Column as Column2

        assert Column2 != sqlf_column

        from pyspark.sql.dataframe import DataFrame as DataFrame2
        from pyspark.sql.dataframe import DataFrameNaFunctions as DataFrameNaFunctions2
        from pyspark.sql.dataframe import (
            DataFrameStatFunctions as DataFrameStatFunctions2,
        )

        assert DataFrame2 != sqlf_dataframe
        assert DataFrameNaFunctions2 != sqlf_dataframe_na_functions
        assert DataFrameStatFunctions2 != sqlf_dataframe_stat_functions

        from pyspark.sql.group import GroupedData as GroupedData2

        assert GroupedData2 != sqlf_grouped_data

        from pyspark.sql.window import WindowSpec as WindowSpec2

        assert WindowSpec2 != sqlf_window_spec

        from pyspark.sql.readwriter import DataFrameReader as DataFrameReader2
        from pyspark.sql.readwriter import DataFrameWriter as DataFrameWriter2

        assert DataFrameReader2 != sqlf_dataframe_reader
        assert DataFrameWriter2 != sqlf_dataframe_writer

        from pyspark.sql.window import Window, WindowSpec

        assert Window != sqlf_window
        assert WindowSpec != sqlf_window_spec

        from pyspark.sql import functions as F
        from pyspark.sql import types

        assert F != sqlf_functions
        assert types != sqlf_types
        assert types.Row != sqlf_row

        from pyspark.sql import UDFRegistration

        assert UDFRegistration != sqlf_udf_registration

        import pyspark.sql.functions as F
        import pyspark.sql.types as types

        assert F != sqlf_functions
        assert types != sqlf_types

    return _check_pyspark_imports



import pytest

from sqlframe import activate
from sqlframe.bigquery import (
    BigQueryCatalog,
    BigQueryDataFrame,
    BigQueryDataFrameNaFunctions,
    BigQueryDataFrameReader,
    BigQueryDataFrameStatFunctions,
    BigQueryDataFrameWriter,
    BigQueryGroupedData,
    BigQuerySession,
    BigQueryUDFRegistration,
)
from sqlframe.bigquery import Column as BigQueryColumn
from sqlframe.bigquery import Row as BigQueryRow
from sqlframe.bigquery import Window as BigQueryWindow
from sqlframe.bigquery import WindowSpec as BigQueryWindowSpec
from sqlframe.bigquery import functions as BigQueryF
from sqlframe.bigquery import types as BigQueryTypes


@pytest.mark.forked
def test_activate_bigquery(check_pyspark_imports):
    check_pyspark_imports(
        "bigquery",
        sqlf_session=BigQuerySession,
        sqlf_catalog=BigQueryCatalog,
        sqlf_column=BigQueryColumn,
        sqlf_dataframe=BigQueryDataFrame,
        sqlf_grouped_data=BigQueryGroupedData,
        sqlf_window=BigQueryWindow,
        sqlf_window_spec=BigQueryWindowSpec,
        sqlf_functions=BigQueryF,
        sqlf_types=BigQueryTypes,
        sqlf_udf_registration=BigQueryUDFRegistration,
        sqlf_dataframe_reader=BigQueryDataFrameReader,
        sqlf_dataframe_writer=BigQueryDataFrameWriter,
        sqlf_dataframe_na_functions=BigQueryDataFrameNaFunctions,
        sqlf_dataframe_stat_functions=BigQueryDataFrameStatFunctions,
        sqlf_row=BigQueryRow,
    )


@pytest.mark.forked
def test_activate_bigquery_default_dataset():
    activate("bigquery", config={"default_dataset": "sqlframe.sqlframe_test"})
    from pyspark.sql import SparkSession

    assert SparkSession == BigQuerySession
    spark = SparkSession.builder.appName("test").getOrCreate()
    assert spark.default_dataset == "sqlframe.sqlframe_test"



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import fnmatch
import json
import typing as t

from sqlglot import exp, parse_one

from sqlframe.base.catalog import Function, _BaseCatalog
from sqlframe.base.mixins.catalog_mixins import (
    GetCurrentCatalogFromFunctionMixin,
    GetCurrentDatabaseFromFunctionMixin,
    ListCatalogsFromInfoSchemaMixin,
    ListColumnsFromInfoSchemaMixin,
    ListDatabasesFromInfoSchemaMixin,
    ListTablesFromInfoSchemaMixin,
    SetCurrentCatalogFromUseMixin,
    SetCurrentDatabaseFromUseMixin,
)
from sqlframe.base.util import normalize_string, schema_, to_schema

if t.TYPE_CHECKING:
    from sqlframe.snowflake.session import SnowflakeSession  # noqa
    from sqlframe.snowflake.dataframe import SnowflakeDataFrame  # noqa


class SnowflakeCatalog(
    SetCurrentCatalogFromUseMixin["SnowflakeSession", "SnowflakeDataFrame"],
    GetCurrentCatalogFromFunctionMixin["SnowflakeSession", "SnowflakeDataFrame"],
    GetCurrentDatabaseFromFunctionMixin["SnowflakeSession", "SnowflakeDataFrame"],
    ListDatabasesFromInfoSchemaMixin["SnowflakeSession", "SnowflakeDataFrame"],
    ListCatalogsFromInfoSchemaMixin["SnowflakeSession", "SnowflakeDataFrame"],
    SetCurrentDatabaseFromUseMixin["SnowflakeSession", "SnowflakeDataFrame"],
    ListTablesFromInfoSchemaMixin["SnowflakeSession", "SnowflakeDataFrame"],
    ListColumnsFromInfoSchemaMixin["SnowflakeSession", "SnowflakeDataFrame"],
    _BaseCatalog["SnowflakeSession", "SnowflakeDataFrame"],
):
    CURRENT_CATALOG_EXPRESSION: exp.Expression = exp.func("current_database")
    UPPERCASE_INFO_SCHEMA = True

    def listFunctions(
        self, dbName: t.Optional[str] = None, pattern: t.Optional[str] = None
    ) -> t.List[Function]:
        """
        Returns a t.List of functions registered in the specified database.

        .. versionadded:: 3.4.0

        Parameters
        ----------
        dbName : str
            name of the database to t.List the functions.
            ``dbName`` can be qualified with catalog name.
        pattern : str
            The pattern that the function name needs to match.

            .. versionchanged: 3.5.0
                Adds ``pattern`` argument.

        Returns
        -------
        t.List
            A t.List of :class:`Function`.

        Notes
        -----
        If no database is specified, the current database and catalog
        are used. This API includes all temporary functions.

        Examples
        --------
        >>> spark.catalog.t.listFunctions()
        [Function(name=...

        >>> spark.catalog.t.listFunctions(pattern="to_*")
        [Function(name=...

        >>> spark.catalog.t.listFunctions(pattern="*not_existing_func*")
        []
        """
        if dbName is None:
            schema = schema_(
                db=exp.parse_identifier(
                    self.currentDatabase(), dialect=self.session.output_dialect
                ),
                catalog=exp.parse_identifier(
                    self.currentCatalog(), dialect=self.session.output_dialect
                ),
            )
        else:
            dbName = normalize_string(dbName, from_dialect="input", is_schema=True)
            schema = to_schema(dbName, dialect=self.session.input_dialect)
            if not schema.catalog:
                schema.set(
                    "catalog",
                    exp.parse_identifier(
                        self.currentCatalog(), dialect=self.session.output_dialect
                    ),
                )
        query = parse_one(
            f"""SHOW USER FUNCTIONS IN {schema.sql(dialect=self.session.input_dialect)}""",
            dialect=self.session.input_dialect,
        )
        functions = [
            Function(
                name=normalize_string(x["name"], from_dialect="execution", to_dialect="output"),
                catalog=normalize_string(
                    x["catalog_name"], from_dialect="execution", to_dialect="output"
                ),
                namespace=[
                    normalize_string(
                        x["schema_name"], from_dialect="execution", to_dialect="output"
                    )
                ],
                description=None,
                className="",
                isTemporary=False,
            )
            for x in self.session._collect(query)
        ]
        if pattern:
            normalized_pattern = normalize_string(
                pattern, from_dialect="input", to_dialect="output", is_pattern=True
            )
            functions = [x for x in functions if fnmatch.fnmatch(x.name, normalized_pattern)]
        return functions

    def get_columns(self, table: exp.Table | str) -> t.Dict[str, exp.DataType]:
        table = (
            normalize_string(table, from_dialect="input", is_table=True)
            if isinstance(table, str)
            else table
        )
        table = exp.to_table(table, dialect=self.session.input_dialect)
        if not table.catalog:
            table.set(
                "catalog",
                exp.parse_identifier(
                    normalize_string(
                        self.currentCatalog(), from_dialect="output", to_dialect="input"
                    ),
                    dialect=self.session.input_dialect,
                ),
            )
        if not table.db:
            table.set(
                "db",
                exp.parse_identifier(
                    normalize_string(
                        self.currentDatabase(), from_dialect="output", to_dialect="input"
                    ),
                    dialect=self.session.input_dialect,
                ),
            )
        sql = f"SHOW COLUMNS IN TABLE {table.sql(dialect=self.session.input_dialect)}"
        results = self.session._collect(sql)
        return {
            normalize_string(
                row["column_name"],
                from_dialect="execution",
                to_dialect="output",
                is_column=True,
            ): exp.DataType.build(
                normalize_string(
                    json.loads(row["data_type"])["type"],
                    from_dialect="execution",
                    to_dialect="output",
                    is_datatype=True,
                ),
                dialect=self.session.output_dialect,
                udt=True,
            )
            for row in results
        }



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import sys
import typing as t

from sqlglot import expressions as exp
from sqlglot.helper import flatten

from sqlframe.base import functions as F

if t.TYPE_CHECKING:
    from sqlframe.base._typing import ColumnOrName


class Window:
    _JAVA_MIN_LONG = -(1 << 63)  # -9223372036854775808
    _JAVA_MAX_LONG = (1 << 63) - 1  # 9223372036854775807
    _PRECEDING_THRESHOLD = max(-sys.maxsize, _JAVA_MIN_LONG)
    _FOLLOWING_THRESHOLD = min(sys.maxsize, _JAVA_MAX_LONG)

    unboundedPreceding: int = _JAVA_MIN_LONG

    unboundedFollowing: int = _JAVA_MAX_LONG

    currentRow: int = 0

    @classmethod
    def partitionBy(cls, *cols: t.Union[ColumnOrName, t.List[ColumnOrName]]) -> WindowSpec:
        return WindowSpec().partitionBy(*cols)

    @classmethod
    def orderBy(cls, *cols: t.Union[ColumnOrName, t.List[ColumnOrName]]) -> WindowSpec:
        return WindowSpec().orderBy(*cols)

    @classmethod
    def rowsBetween(cls, start: int, end: int) -> WindowSpec:
        return WindowSpec().rowsBetween(start, end)

    @classmethod
    def rangeBetween(cls, start: int, end: int) -> WindowSpec:
        return WindowSpec().rangeBetween(start, end)


class WindowSpec:
    def __init__(self, expression: exp.Expression = exp.Window()):
        self.expression = expression

    def copy(self):
        return WindowSpec(self.expression.copy())

    def sql(self, **kwargs) -> str:
        from sqlframe.base.session import _BaseSession

        return self.expression.sql(dialect=_BaseSession().input_dialect, **kwargs)

    def partitionBy(self, *cols: t.Union[ColumnOrName, t.List[ColumnOrName]]) -> WindowSpec:
        from sqlframe.base.column import Column

        cols = flatten(cols) if isinstance(cols[0], (list, set, tuple)) else cols  # type: ignore
        expressions = [Column.ensure_col(x).expression for x in cols]  # type: ignore
        window_spec = self.copy()
        partition_by_expressions = window_spec.expression.args.get("partition_by", [])
        partition_by_expressions.extend(expressions)
        window_spec.expression.set("partition_by", partition_by_expressions)
        return window_spec

    def orderBy(self, *cols: t.Union[ColumnOrName, t.List[ColumnOrName]]) -> WindowSpec:
        from sqlframe.base.column import Column

        cols = flatten(cols) if isinstance(cols[0], (list, set, tuple)) else cols  # type: ignore
        expressions = [Column.ensure_col(x).expression for x in cols]  # type: ignore
        window_spec = self.copy()
        if window_spec.expression.args.get("order") is None:
            window_spec.expression.set("order", exp.Order(expressions=[]))
        order_by = window_spec.expression.args["order"].expressions
        order_by.extend(expressions)
        window_spec.expression.args["order"].set("expressions", order_by)
        return window_spec

    def _calc_start_end(
        self, start: int, end: int
    ) -> t.Dict[str, t.Optional[t.Union[str, exp.Expression]]]:
        kwargs: t.Dict[str, t.Optional[t.Union[str, exp.Expression]]] = {
            "start_side": None,
            "end_side": None,
        }
        if start == Window.currentRow:
            kwargs["start"] = "CURRENT ROW"
        else:
            kwargs = {
                **kwargs,
                **{
                    "start_side": "PRECEDING",
                    "start": (
                        "UNBOUNDED"
                        if start <= Window.unboundedPreceding
                        else F.lit(start).expression
                    ),
                },
            }
        if end == Window.currentRow:
            kwargs["end"] = "CURRENT ROW"
        else:
            kwargs = {
                **kwargs,
                **{
                    "end_side": "FOLLOWING",
                    "end": (
                        "UNBOUNDED" if end >= Window.unboundedFollowing else F.lit(end).expression
                    ),
                },
            }
        return kwargs

    def rowsBetween(self, start: int, end: int) -> WindowSpec:
        window_spec = self.copy()
        spec = self._calc_start_end(start, end)
        spec["kind"] = "ROWS"
        window_spec.expression.set(
            "spec",
            exp.WindowSpec(
                **{**window_spec.expression.args.get("spec", exp.WindowSpec()).args, **spec}
            ),
        )
        return window_spec

    def rangeBetween(self, start: int, end: int) -> WindowSpec:
        window_spec = self.copy()
        spec = self._calc_start_end(start, end)
        spec["kind"] = "RANGE"
        window_spec.expression.set(
            "spec",
            exp.WindowSpec(
                **{**window_spec.expression.args.get("spec", exp.WindowSpec()).args, **spec}
            ),
        )
        return window_spec



from __future__ import annotations

import logging
import math
import re
import typing as t

from sqlglot import exp as expression
from sqlglot.dialects.dialect import build_formatted_time
from sqlglot.helper import ensure_list
from sqlglot.helper import flatten as _flatten

from sqlframe.base.column import Column
from sqlframe.base.util import (
    get_func_from_session,
)

if t.TYPE_CHECKING:
    from sqlframe.base._typing import ColumnOrLiteral, ColumnOrName

logger = logging.getLogger(__name__)


def e_literal() -> Column:
    lit = get_func_from_session("lit")

    return lit(math.e)


def expm1_from_exp(col: ColumnOrName) -> Column:
    exp = get_func_from_session("exp")
    lit = get_func_from_session("lit")
    return exp(col) - lit(1)


def log1p_from_log(col: ColumnOrName) -> Column:
    from sqlframe.base.session import _BaseSession

    session: _BaseSession = _BaseSession()
    log = get_func_from_session("log", session)
    lit = get_func_from_session("lit", session)
    return log(col + lit(1))


def rint_from_round(col: ColumnOrName) -> Column:
    from sqlframe.base.session import _BaseSession

    round = get_func_from_session("round", _BaseSession())
    return round(col, 0)


def kurtosis_from_kurtosis_pop(col: ColumnOrName) -> Column:
    return Column.invoke_anonymous_function(col, "KURTOSIS_POP")


def collect_set_from_list_distinct(col: ColumnOrName) -> Column:
    collect_list = get_func_from_session("collect_list")
    return collect_list(Column(expression.Distinct(expressions=[Column(col).expression])))


def first_always_ignore_nulls(col: ColumnOrName, ignorenulls: t.Optional[bool] = None) -> Column:
    from sqlframe.base.functions import first

    return first(col)


def to_timestamp_with_time_zone(col: ColumnOrName, format: t.Optional[str] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    if format is not None:
        return Column.invoke_expression_over_column(
            col, expression.StrToTime, format=_BaseSession().format_time(format)
        )

    return Column.ensure_col(col).cast("timestamp with time zone", dialect="postgres")


def to_timestamp_tz(col: ColumnOrName, format: t.Optional[str] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    if format is not None:
        return Column.invoke_expression_over_column(
            col, expression.StrToTime, format=_BaseSession().format_time(format)
        )

    return Column.ensure_col(col).cast("timestamptz", dialect="duckdb")


def to_timestamp_just_timestamp(col: ColumnOrName, format: t.Optional[str] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    if format is not None:
        return Column.invoke_expression_over_column(
            col, expression.StrToTime, format=_BaseSession().format_time(format)
        )

    return Column.ensure_col(col).cast("datetime", dialect="bigquery")


def bitwise_not_from_bitnot(col: ColumnOrName) -> Column:
    return Column.invoke_anonymous_function(col, "BITNOT")


def factorial_from_case_statement(col: ColumnOrName) -> Column:
    from sqlframe.base.session import _BaseSession

    session: _BaseSession = _BaseSession()
    when = get_func_from_session("when", session)
    col_func = get_func_from_session("col", session)
    lit = get_func_from_session("lit", session)
    return (
        when(
            col_func(col) == lit(1),
            lit(1),
        )
        .when(
            col_func(col) == lit(2),
            lit(2),
        )
        .when(
            col_func(col) == lit(3),
            lit(6),
        )
        .when(
            col_func(col) == lit(4),
            lit(24),
        )
        .when(
            col_func(col) == lit(5),
            lit(120),
        )
        .when(
            col_func(col) == lit(6),
            lit(720),
        )
        .when(
            col_func(col) == lit(7),
            lit(5040),
        )
        .when(
            col_func(col) == lit(8),
            lit(40320),
        )
        .when(
            col_func(col) == lit(9),
            lit(362880),
        )
        .when(
            col_func(col) == lit(10),
            lit(3628800),
        )
        .when(
            col_func(col) == lit(11),
            lit(39916800),
        )
        .when(
            col_func(col) == lit(12),
            lit(479001600),
        )
        .when(
            col_func(col) == lit(13),
            lit(6227020800),
        )
        .when(
            col_func(col) == lit(14),
            lit(87178291200),
        )
        .when(
            col_func(col) == lit(15),
            lit(1307674368000),
        )
        .when(
            col_func(col) == lit(16),
            lit(20922789888000),
        )
        .when(
            col_func(col) == lit(17),
            lit(355687428096000),
        )
        .when(
            col_func(col) == lit(18),
            lit(6402373705728000),
        )
        .when(
            col_func(col) == lit(19),
            lit(121645100408832000),
        )
        .when(
            col_func(col) == lit(20),
            lit(2432902008176640000),
        )
        .otherwise(
            lit(None),
        )
    )


def factorial_ensure_int(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column.invoke_anonymous_function(col_func(col).cast("integer"), "FACTORIAL")


def skewness_from_skew(col: ColumnOrName) -> Column:
    return Column.invoke_anonymous_function(col, "SKEW")


def isnan_using_equal(col: ColumnOrName) -> Column:
    lit = get_func_from_session("lit")
    return Column(
        expression.EQ(this=Column(col).expression, expression=lit(float("nan")).expression)
    )


def isnull_using_equal(col: ColumnOrName) -> Column:
    lit = get_func_from_session("lit")
    col_func = get_func_from_session("col")
    return Column(expression.Is(this=col_func(col).expression, expression=lit(None).expression))


def nanvl_as_case(col1: ColumnOrName, col2: ColumnOrName) -> Column:
    when = get_func_from_session("when")
    isnan = get_func_from_session("isnan")
    col = get_func_from_session("col")
    return when(~isnan(col1), col(col1)).otherwise(col(col2))


def percentile_approx_without_accuracy(
    col: ColumnOrName,
    percentage: t.Union[ColumnOrLiteral, t.List[float], t.Tuple[float]],
    accuracy: t.Optional[float] = None,
) -> Column:
    from sqlframe.base.functions import percentile_approx

    if accuracy:
        logger.warning("Accuracy is ignored since it is not supported in this dialect")
    return percentile_approx(col, percentage)


def percentile_approx_without_accuracy_and_plural(
    col: ColumnOrName,
    percentage: t.Union[ColumnOrLiteral, t.List[float], t.Tuple[float]],
    accuracy: t.Optional[float] = None,
) -> Column:
    lit = get_func_from_session("lit")
    array = get_func_from_session("array")
    col_func = get_func_from_session("col")

    def make_bracket_approx_percentile(percentage: float) -> expression.Bracket:
        return expression.Bracket(
            this=expression.Anonymous(
                this="APPROX_QUANTILES",
                expressions=[col_func(col).expression, lit(100).expression],
            ),
            expressions=[lit(int(percentage * 100)).cast("int").expression],
            offset=0,
            safe=False,
        )

    if accuracy:
        logger.warning("Accuracy is ignored since it is not supported in this dialect")
    if isinstance(percentage, (list, tuple)):
        return array(*[make_bracket_approx_percentile(p) for p in percentage])
    return Column(make_bracket_approx_percentile(percentage))  # type: ignore


def percentile_approx_without_accuracy_and_max_array(
    col: ColumnOrName,
    percentage: t.Union[ColumnOrLiteral, t.List[float], t.Tuple[float]],
    accuracy: t.Optional[float] = None,
) -> Column:
    from sqlframe.base.functions import percentile_approx

    lit = get_func_from_session("lit")
    array = get_func_from_session("array")
    col_func = get_func_from_session("col")

    def make_approx_percentile(percentage: float) -> expression.Anonymous:
        return expression.Anonymous(
            this="APPROX_PERCENTILE",
            expressions=[col_func(col).expression, lit(percentage).expression],
        )

    if accuracy:
        logger.warning("Accuracy is ignored since it is not supported in this dialect")
    if isinstance(percentage, (list, tuple)):
        return array(*[make_approx_percentile(p) for p in percentage])
    return percentile_approx(col, percentage)


def percentile_without_disc(
    col: ColumnOrName,
    percentage: t.Union[ColumnOrLiteral, t.List[float], t.Tuple[float]],
    frequency: t.Optional[ColumnOrLiteral] = None,
) -> Column:
    lit = get_func_from_session("lit")
    col_func = get_func_from_session("col")

    percentage_col = percentage if isinstance(percentage, Column) else lit(percentage)
    func_expressions = [
        col_func(col).expression,
        percentage_col.expression,
    ]
    if frequency:
        func_expressions.append(frequency if isinstance(frequency, Column) else lit(frequency))
    return Column(
        expression.Anonymous(
            this="PERCENTILE",
            expressions=func_expressions,
        )
    )


def rand_no_seed(seed: t.Optional[ColumnOrLiteral] = None) -> Column:
    from sqlframe.base.functions import rand

    if seed:
        logger.warning("Seed is ignored since it is not supported in this dialect")
    return rand()


def round_cast_as_numeric(col: ColumnOrName, scale: t.Optional[int] = None) -> Column:
    from sqlframe.base.functions import round

    col_func = get_func_from_session("col")

    return round(col_func(col).cast("numeric"), scale)


def bround_using_half_even(col: ColumnOrName, scale: t.Optional[int] = None) -> Column:
    lit_func = get_func_from_session("lit")

    return Column.invoke_anonymous_function(col, "ROUND", scale, lit_func("HALF_TO_EVEN"))  # type: ignore


def shiftleft_from_bitshiftleft(col: ColumnOrName, numBits: int) -> Column:
    col_func = get_func_from_session("col")
    lit = get_func_from_session("lit")

    return Column(
        expression.Anonymous(
            this="BITSHIFTLEFT",
            expressions=[col_func(col).expression, lit(numBits).expression],
        )
    )


def shiftright_from_bitshiftright(col: ColumnOrName, numBits: int) -> Column:
    col_func = get_func_from_session("col")
    lit = get_func_from_session("lit")

    return Column(
        expression.Anonymous(
            this="BITSHIFTRIGHT",
            expressions=[col_func(col).expression, lit(numBits).expression],
        )
    )


def struct_with_eq(
    col: t.Union[ColumnOrName, t.Iterable[ColumnOrName]], *cols: ColumnOrName
) -> Column:
    from sqlframe.base.session import _BaseSession

    col_func = get_func_from_session("col")

    columns = [col_func(x) for x in ensure_list(col) + list(cols)]
    expressions = []
    for column in columns:
        expressions.append(
            expression.PropertyEQ(
                this=expression.parse_identifier(
                    column.alias_or_name, dialect=_BaseSession().input_dialect
                ),
                expression=column.expression,
            )
        )
    return Column(expression.Struct(expressions=expressions))


def year_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="year"), expression=col_func(col).cast("date").expression
        )
    )


def quarter_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="quarter"), expression=col_func(col).cast("date").expression
        )
    )


def month_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="month"), expression=col_func(col).cast("date").expression
        )
    )


def dayofweek_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="dayofweek"), expression=col_func(col).cast("date").expression
        )
    )


def dayofweek_from_extract_with_isodow(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="isodow"), expression=col_func(col).cast("date").expression
        )
    )


def dayofmonth_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="dayofmonth"), expression=col_func(col).cast("date").expression
        )
    )


def dayofmonth_from_extract_with_day(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="day"), expression=col_func(col).cast("date").expression
        )
    )


def dayofyear_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="dayofyear"), expression=col_func(col).cast("date").expression
        )
    )


def dayofyear_from_extract_doy(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="doy"), expression=col_func(col).cast("date").expression
        )
    )


def hour_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(this=expression.Var(this="hour"), expression=col_func(col).expression)
    )


def minute_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(this=expression.Var(this="minute"), expression=col_func(col).expression)
    )


def second_from_extract(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(this=expression.Var(this="second"), expression=col_func(col).expression)
    )


def weekofyear_from_extract_as_week(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="week"), expression=col_func(col).cast("date").expression
        )
    )


def weekofyear_from_extract_as_isoweek(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Extract(
            this=expression.Var(this="ISOWEEK"), expression=col_func(col).cast("date").expression
        )
    )


def make_date_casted_as_integer(
    year: ColumnOrName, month: ColumnOrName, day: ColumnOrName
) -> Column:
    from sqlframe.base.functions import make_date

    col_func = get_func_from_session("col")

    return make_date(
        col_func(year).cast("integer"),
        col_func(month).cast("integer"),
        col_func(day).cast("integer"),
    )


def make_date_from_date_func(year: ColumnOrName, month: ColumnOrName, day: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="DATE",
            expressions=[
                col_func(year).cast("integer").expression,
                col_func(month).cast("integer").expression,
                col_func(day).cast("integer").expression,
            ],
        )
    )


def make_date_date_from_parts(year: ColumnOrName, month: ColumnOrName, day: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="DATE_FROM_PARTS",
            expressions=[
                col_func(year).cast("integer").expression,
                col_func(month).cast("integer").expression,
                col_func(day).cast("integer").expression,
            ],
        )
    )


def date_add_no_date_sub(
    col: ColumnOrName, days: t.Union[ColumnOrName, int], cast_as_date: bool = True
) -> Column:
    lit_func = get_func_from_session("col")

    if isinstance(days, int):
        days = lit_func(days)

    result = Column.invoke_expression_over_column(
        Column.ensure_col(col).cast("date"),
        expression.DateAdd,
        expression=days,
        unit=expression.Var(this="DAY"),
    )
    if cast_as_date:
        return result.cast("date")
    return result


def date_sub_by_date_add(
    col: ColumnOrName, days: t.Union[ColumnOrName, int], cast_as_date: bool = True
) -> Column:
    lit_func = get_func_from_session("col")
    date_add_func = get_func_from_session("date_add")

    return date_add_func(col, days * lit_func(-1), cast_as_date)


def to_date_from_timestamp(col: ColumnOrName, format: t.Optional[str] = None) -> Column:
    from sqlframe.base.functions import to_date

    to_timestamp = get_func_from_session("to_timestamp")

    return to_date(to_timestamp(col, format))


def to_date_time_format(col: ColumnOrName, format: t.Optional[str] = None) -> Column:
    from sqlframe.base.functions import to_date
    from sqlframe.base.session import _BaseSession

    return to_date(col, format=format or _BaseSession().default_time_format)


def last_day_with_cast(col: ColumnOrName) -> Column:
    from sqlframe.base.functions import last_day

    col_func = get_func_from_session("col")

    return last_day(col_func(col).cast("date"))


def sha1_force_sha1_and_to_hex(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="TO_HEX",
            expressions=[
                expression.Anonymous(
                    this="SHA1",
                    expressions=[col_func(col).expression],
                )
            ],
        )
    )


def hash_from_farm_fingerprint(*cols: ColumnOrName) -> Column:
    if len(cols) > 1:
        raise ValueError("This dialect only supports a single column for calculating hash")

    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="FARM_FINGERPRINT",
            expressions=[col_func(cols[0]).expression],
        )
    )


def date_add_by_multiplication(
    col: ColumnOrName, days: t.Union[ColumnOrName, int], cast_as_date: bool = True
) -> Column:
    from sqlframe.base.functions import date_add

    col_func = get_func_from_session("col")

    if isinstance(days, int):
        value = date_add(col, days)
    else:
        value = date_add(col, 1, cast_as_date=False) * col_func(days)
    if cast_as_date:
        return value.cast("date")
    return value


def date_sub_by_multiplication(
    col: ColumnOrName, days: t.Union[ColumnOrName, int], cast_as_date: bool = True
) -> Column:
    from sqlframe.base.functions import date_sub

    col_func = get_func_from_session("col")

    if isinstance(days, int):
        value = date_sub(col, days)
    else:
        value = date_sub(col, 1, cast_as_date=False) * col_func(days)
    if cast_as_date:
        return value.cast("date")
    return value


def date_diff_with_subtraction(end: ColumnOrName, start: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return col_func(end).cast("date") - col_func(start).cast("date")


def add_months_by_multiplication(
    start: ColumnOrName, months: t.Union[ColumnOrName, int], cast_as_date: bool = True
) -> Column:
    from sqlframe.base.functions import add_months

    col_func = get_func_from_session("col")
    lit = get_func_from_session("lit")

    multiple_value = lit(months) if isinstance(months, int) else col_func(months)
    value = col_func(add_months(start, 1, cast_as_date=False).expression.unnest()) * multiple_value
    if cast_as_date:
        return value.cast("date")
    return value


def add_months_using_func(
    start: ColumnOrName, months: t.Union[ColumnOrName, int], cast_as_date: bool = True
) -> Column:
    from sqlframe.base.functions import add_months

    if isinstance(months, int):
        months = get_func_from_session("lit")(months)
    else:
        months = Column.ensure_col(months)

    value = Column(
        expression.Anonymous(
            this="ADD_MONTHS",
            expressions=[
                Column.ensure_col(start).expression,
                months.expression,  # type: ignore
            ],
        )
    )

    if cast_as_date:
        return value.cast("date")
    return value


def months_between_from_age_and_extract(
    date1: ColumnOrName, date2: ColumnOrName, roundOff: t.Optional[bool] = None
) -> Column:
    lit = get_func_from_session("lit")
    col_func = get_func_from_session("col")

    if roundOff:
        logger.warning("Round off is ignored since it is not supported in this dialect")
    age_expression = expression.Anonymous(
        this="AGE",
        expressions=[
            col_func(date1).cast("date").expression,
            col_func(date2).cast("date").expression,
        ],
    )
    return (
        Column(
            expression.Extract(this=expression.Var(this="year"), expression=age_expression)
            * expression.Literal.number(12)
        )
        + Column(expression.Extract(this=expression.Var(this="month"), expression=age_expression))
        + lit(1)
    ).cast("bigint")


def months_between_cast_as_date_cast_roundoff(
    date1: ColumnOrName, date2: ColumnOrName, roundOff: t.Optional[bool] = None
) -> Column:
    from sqlframe.base.functions import months_between

    col_func = get_func_from_session("col")

    date1 = col_func(date1).cast("date")
    date2 = col_func(date2).cast("date")

    value = months_between(date1, date2)

    if roundOff:
        return value.cast("bigint")
    return value


def from_unixtime_from_timestamp(col: ColumnOrName, format: t.Optional[str] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    col_func = get_func_from_session("col")

    return Column.invoke_expression_over_column(
        Column(
            expression.Anonymous(
                this="TO_TIMESTAMP",
                expressions=[col_func(col).expression],
            )
        ),
        expression.TimeToStr,
        format=_BaseSession().format_time(format),
    )


def unix_timestamp_from_extract(
    timestamp: t.Optional[ColumnOrName] = None, format: t.Optional[str] = None
) -> Column:
    to_timestamp = get_func_from_session("to_timestamp")

    return Column(
        expression.Extract(
            this=expression.Var(this="epoch"), expression=to_timestamp(timestamp, format).expression
        )
    ).cast("bigint")


def base64_from_blob(col: ColumnOrLiteral) -> Column:
    return Column.invoke_expression_over_column(Column(col).cast("blob"), expression.ToBase64)


def bas64_from_encode(col: ColumnOrLiteral) -> Column:
    return Column(
        expression.Encode(
            this=Column(col).cast("bytea").expression, charset=expression.Literal.string("base64")
        )
    )


def base64_from_base64_encode(col: ColumnOrLiteral) -> Column:
    return Column(
        expression.Anonymous(
            this="BASE64_ENCODE",
            expressions=[Column(col).expression],
        )
    )


def unbase64_from_decode(col: ColumnOrLiteral) -> Column:
    return Column(
        expression.Decode(this=Column(col).expression, charset=expression.Literal.string("base64"))
    )


def unbase64_from_base64_decode_string(col: ColumnOrLiteral) -> Column:
    return Column(
        expression.Anonymous(
            this="BASE64_DECODE_STRING",
            expressions=[Column(col).expression],
        )
    )


def decode_from_blob(col: ColumnOrLiteral, charset: str) -> Column:
    return Column(
        expression.Decode(
            this=Column(col).cast("blob").expression, charset=expression.Literal.string(charset)
        )
    )


def decode_from_convert_from(col: ColumnOrLiteral, charset: str) -> Column:
    return Column(
        expression.Anonymous(
            this="CONVERT_FROM",
            expressions=[Column(col).cast("bytea").expression, expression.Literal.string(charset)],
        )
    )


def encode_from_convert_to(col: ColumnOrName, charset: str) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="CONVERT_TO",
            expressions=[col_func(col).expression, expression.Literal.string(charset)],
        )
    )


def concat_ws_from_array_to_string(sep: str, *cols: ColumnOrName) -> Column:
    array = get_func_from_session("array")
    lit = get_func_from_session("lit")

    return Column(
        expression.Anonymous(
            this="ARRAY_TO_STRING",
            expressions=[array(*cols).expression, lit(sep).expression],
        )
    )


def format_number_from_to_char(col: ColumnOrName, d: int) -> Column:
    round = get_func_from_session("round")
    format = "FM" + ("999," * 5) + "990" + "D" + ("0" * d)

    return Column(
        expression.ToChar(this=round(col, d).expression, format=expression.Literal.string(format))
    )


def format_string_with_format(format: str, *cols: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="FORMAT",
            expressions=[
                expression.Literal.string(format.replace("%d", "%s")),
                *[col_func(x).cast("string").expression for x in ensure_list(cols)],
            ],
        )
    )


def format_string_with_pipes(format: str, *cols: ColumnOrName) -> Column:
    lit = get_func_from_session("lit")
    col_func = get_func_from_session("col")

    values = format.replace("%d", "%s").split("%s")
    if len(values) != len(cols) + 1:
        raise ValueError("Number of values and columns do not match")
    result = expression.DPipe(
        this=lit(values[0]).expression, expression=col_func(cols[0]).expression
    )
    for i, value in enumerate(values[1:], start=1):
        if i == len(cols):
            result = expression.DPipe(this=result, expression=lit(value).expression)
        else:
            result = expression.DPipe(
                this=expression.DPipe(this=result, expression=lit(value).expression),
                expression=col_func(cols[i]).expression,
            )
    return Column(result)


def instr_using_strpos(col: ColumnOrName, substr: str) -> Column:
    lit = get_func_from_session("lit")
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="STRPOS",
            expressions=[col_func(col).expression, lit(substr).expression],
        )
    )


def overlay_from_substr(
    src: ColumnOrName,
    replace: ColumnOrName,
    pos: t.Union[ColumnOrName, int],
    len: t.Optional[t.Union[ColumnOrName, int]] = None,
) -> Column:
    col_func = get_func_from_session("col")
    lit = get_func_from_session("lit")
    substring = get_func_from_session("substring")
    length_func = get_func_from_session("length")
    length_value = len if len is not None else length_func(replace)
    return Column(
        expression.Concat(
            expressions=[
                substring(col_func(src), 1, col_func(pos) - lit(1)).expression,
                col_func(replace).expression,
                substring(
                    col_func(src), col_func(pos) + col_func(length_value), length_func(src)
                ).expression,
            ]
        )
    )


def levenshtein_edit_distance(
    left: ColumnOrName, right: ColumnOrName, threshold: t.Optional[int] = None
) -> Column:
    if threshold is not None:
        logger.warning("Threshold is ignored since it is not supported in this dialect")
    return Column(
        expression.Anonymous(
            this="EDITDISTANCE",
            expressions=[Column.ensure_col(left).expression, Column.ensure_col(right).expression],
        )
    )


def split_no_limit(str: ColumnOrName, pattern: str, limit: t.Optional[int] = None) -> Column:
    from sqlframe.base.functions import split

    col_func = get_func_from_session("col")

    if limit is not None:
        logger.warning("Limit is ignored since it is not supported in this dialect")
    return split(col_func(str), pattern)


def split_from_regex_split_to_array(
    str: ColumnOrName, pattern: str, limit: t.Optional[int] = None
) -> Column:
    col_func = get_func_from_session("col")

    if limit is not None:
        logger.warning("Limit is ignored since it is not supported in this dialect")
    return Column(
        expression.Anonymous(
            this="REGEXP_SPLIT_TO_ARRAY",
            expressions=[
                col_func(str).expression,
                expression.Literal.string(pattern),
            ],
        )
    )


def split_with_split(str: ColumnOrName, pattern: str, limit: t.Optional[int] = None) -> Column:
    lit = get_func_from_session("lit")
    col_func = get_func_from_session("col")

    if limit is not None:
        logger.warning("Limit is ignored since it is not supported in this dialect")
    return Column(
        expression.Anonymous(
            this="SPLIT",
            expressions=[col_func(str).expression, lit(pattern).expression],
        )
    )


def regexp_extract_coalesce_empty_str(
    str: ColumnOrName, pattern: str, idx: t.Optional[int] = None
) -> Column:
    from sqlframe.base.functions import regexp_extract

    coalesce = get_func_from_session("coalesce")
    lit_func = get_func_from_session("lit")

    return coalesce(regexp_extract(str, pattern, idx), lit_func(""))


def array_contains_any(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    lit = get_func_from_session("lit")
    value_col = value if isinstance(value, Column) else lit(value)
    col_func = get_func_from_session("col")

    return Column(
        expression.EQ(
            this=value_col.expression,
            expression=expression.Anonymous(this="ANY", expressions=[col_func(col).expression]),
        )
    )


def arrays_overlap_using_intersect(col1: ColumnOrName, col2: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.GT(
            this=expression.ArraySize(
                this=expression.Anonymous(
                    this="ARRAY_INTERSECT",
                    expressions=[col_func(col1).expression, col_func(col2).expression],
                )
            ),
            expression=expression.Literal.number(0),
        )
    )


def arrays_overlap_renamed(col1: ColumnOrName, col2: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="ARRAYS_OVERLAP",
            expressions=[col_func(col1).expression, col_func(col2).expression],
        )
    )


def slice_as_list_slice(
    x: ColumnOrName, start: t.Union[ColumnOrName, int], length: t.Union[ColumnOrName, int]
) -> Column:
    lit = get_func_from_session("lit")

    start_col = start if isinstance(start, Column) else lit(start)
    length_col = length if isinstance(length, Column) else lit(length)
    return Column.invoke_anonymous_function(x, "LIST_SLICE", start_col, start_col + length_col)


def slice_with_brackets(
    x: ColumnOrName, start: t.Union[ColumnOrName, int], length: t.Union[ColumnOrName, int]
) -> Column:
    lit = get_func_from_session("lit")

    start_col = start if isinstance(start, Column) else lit(start)
    length_col = length if isinstance(length, Column) else lit(length)
    col_func = get_func_from_session("col")

    return Column(
        expression.Bracket(
            this=col_func(x).expression,
            expressions=[
                expression.Slice(
                    this=start_col.expression,
                    expression=(start_col + length_col).expression,
                )
            ],
        )
    )


def array_join_no_null_replacement(
    col: ColumnOrName, delimiter: str, null_replacement: t.Optional[str] = None
) -> Column:
    from sqlframe.base.functions import array_join

    if null_replacement is None:
        logger.warning("Null replacement is ignored since it is not supported in this dialect")
    return array_join(col, delimiter)


def array_join_null_replacement_with_transform(
    col: ColumnOrName, delimiter: str, null_replacement: t.Optional[str] = None
) -> Column:
    from sqlframe.base.functions import array_join

    col_func = get_func_from_session("col")

    if null_replacement is None:
        return array_join(col, delimiter, null_replacement)
    col = Column(
        expression.Anonymous(
            this="LIST_TRANSFORM",
            expressions=[
                col_func(col).expression,
                expression.Lambda(
                    this=expression.Coalesce(
                        this=expression.Cast(
                            this=expression.Identifier(this="x"),
                            to=expression.DataType.build("STRING"),
                        ),
                        expressions=[expression.Literal.string(null_replacement)],
                    ),
                    expressions=[expression.Identifier(this="x")],
                ),
            ],
        )
    )
    return array_join(col, delimiter)


def array_contains_cast_variant(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    from sqlframe.base.functions import array_contains

    lit = get_func_from_session("lit")
    value_col = value if isinstance(value, Column) else lit(value)
    return array_contains(col, value_col.cast("variant"))


def arrays_overlap_as_plural(col1: ColumnOrName, col2: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="ARRAYS_OVERLAP",
            expressions=[col_func(col1).expression, col_func(col2).expression],
        )
    )


def slice_as_array_slice(
    x: ColumnOrName, start: t.Union[ColumnOrName, int], length: t.Union[ColumnOrName, int]
) -> Column:
    lit = get_func_from_session("lit")

    start_col = start if isinstance(start, Column) else lit(start)
    length_col = length if isinstance(length, Column) else lit(length)
    return Column.invoke_anonymous_function(
        x, "ARRAY_SLICE", start_col - lit(1), start_col + length_col
    )


def array_position_cast_variant_and_flip(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    when = get_func_from_session("when")
    lit = get_func_from_session("lit")
    value_col = value if isinstance(value, Column) else lit(value)
    # Some engines return NULL if item is not found but Spark expects 0 so we coalesce to 0
    resp = Column.invoke_anonymous_function(value_col.cast("variant"), "ARRAY_POSITION", col)
    return when(resp.isNotNull(), resp + lit(1)).otherwise(lit(0))


def array_intersect_using_intersection(col1: ColumnOrName, col2: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="ARRAY_INTERSECTION",
            expressions=[col_func(col1).expression, col_func(col2).expression],
        )
    )


def element_at_using_brackets(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    col_func = get_func_from_session("col")
    lit = get_func_from_session("lit")
    # SQLGlot will auto add 1 to whatever we pass in for the brackets even though the value is already 1 based.
    value = value if isinstance(value, Column) else lit(value)
    if [x for x in value.expression.find_all(expression.Literal) if x.is_number]:
        value = value - lit(1)
    return Column(expression.Bracket(this=col_func(col).expression, expressions=[value.expression]))  # type: ignore


def array_remove_using_filter(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    lit = get_func_from_session("lit")
    col_func = get_func_from_session("col")

    value = value if isinstance(value, Column) else lit(value)
    return Column(
        expression.Anonymous(
            this="LIST_FILTER",
            expressions=[
                col_func(col).expression,
                expression.Lambda(
                    this=expression.NEQ(
                        this=expression.Identifier(this="x"), expression=value.expression
                    ),
                    expressions=[expression.Identifier(this="x")],
                ),
            ],
        )
    )


def array_union_using_list_concat(col1: ColumnOrName, col2: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="LIST_DISTINCT",
            expressions=[
                expression.Anonymous(
                    this="LIST_CONCAT",
                    expressions=[col_func(col1).expression, col_func(col2).expression],
                )
            ],
        )
    )


def array_union_using_array_concat(col1: ColumnOrName, col2: ColumnOrName) -> Column:
    array_distinct = get_func_from_session("array_distinct")
    col_func = get_func_from_session("col")

    return array_distinct(
        expression.ArrayConcat(
            this=col_func(col1).expression, expressions=[col_func(col2).expression]
        )
    )


def get_json_object_using_arrow_op(col: ColumnOrName, path: str) -> Column:
    col_func = get_func_from_session("col")
    path = path.replace("$.", "")
    return Column(
        expression.JSONExtract(
            this=expression.Cast(
                this=col_func(col).expression, to=expression.DataType.build("JSON")
            ),
            expression=expression.JSONPath(
                expressions=[expression.JSONPathRoot(), expression.JSONPathKey(this=path)]
            ),
            only_json_types=True,
        )
    )


def get_json_object_cast_object(col: ColumnOrName, path: str) -> Column:
    from sqlframe.base.functions import get_json_object

    col_func = get_func_from_session("col")

    return get_json_object(col_func(col).cast("variant"), path)


def get_json_object_using_function(col: ColumnOrName, path: str) -> Column:
    lit = get_func_from_session("lit")
    return Column.invoke_anonymous_function(col, "GET_JSON_OBJECT", lit(path))


def create_map_with_cast(*cols: t.Union[ColumnOrName, t.Iterable[ColumnOrName]]) -> Column:
    from sqlframe.base.functions import create_map

    col = get_func_from_session("col")

    columns = list(_flatten(cols)) if not isinstance(cols[0], (str, Column)) else cols
    col1_dtype = col(columns[0]).dtype or "VARCHAR"
    col2_dtype = col(columns[1]).dtype or "VARCHAR"
    return create_map(*cols).cast(f"MAP({col1_dtype}, {col2_dtype})")


def array_min_from_sort(col: ColumnOrName) -> Column:
    element_at = get_func_from_session("element_at")
    array_sort = get_func_from_session("array_sort")

    return element_at(array_sort(col), 1)


def array_min_from_subquery(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    explode = get_func_from_session("explode")
    select = expression.Select(
        expressions=[
            expression.Min(
                this=col_func("x").expression,
            )
        ],
    )
    select.set(
        "from",
        expression.From(
            this=explode(col).alias("x").expression,
        ),
    )

    return Column(expression.Subquery(this=select)).alias(col_func(col).alias_or_name)


def array_max_from_sort(col: ColumnOrName) -> Column:
    element_at = get_func_from_session("element_at")
    array_sort = get_func_from_session("array_sort")

    return element_at(array_sort(col), -1)


def array_max_from_subquery(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    explode = get_func_from_session("explode")
    select = expression.Select(
        expressions=[
            expression.Max(
                this=col_func("x").expression,
            )
        ],
    )
    select.set(
        "from",
        expression.From(
            this=explode(col).alias("x").expression,
        ),
    )

    return Column(expression.Subquery(this=select)).alias(col_func(col).alias_or_name)


def sort_array_using_array_sort(col: ColumnOrName, asc: t.Optional[bool] = None) -> Column:
    col_func = get_func_from_session("col")
    lit_func = get_func_from_session("lit")
    expressions = [col_func(col).expression]
    asc = asc if asc is not None else True
    expressions.append(lit_func(asc).expression)
    if asc:
        expressions.append(lit_func(True).expression)
    else:
        expressions.append(lit_func(False).expression)

    return Column(
        expression.Anonymous(
            this="ARRAY_SORT",
            expressions=expressions,
        )
    )


def flatten_using_array_flatten(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="ARRAY_FLATTEN",
            expressions=[col_func(col).expression],
        )
    )


def map_concat_using_map_cat(*cols: t.Union[ColumnOrName, t.Iterable[ColumnOrName]]) -> Column:
    columns = list(flatten(cols)) if not isinstance(cols[0], (str, Column)) else cols  # type: ignore
    if len(columns) == 1:
        return Column.invoke_anonymous_function(columns[0], "MAP_CAT")  # type: ignore
    return Column.invoke_anonymous_function(columns[0], "MAP_CAT", *columns[1:])  # type: ignore


def sequence_from_generate_series(
    start: ColumnOrName, stop: ColumnOrName, step: t.Optional[ColumnOrName] = None
) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="GENERATE_SERIES",
            expressions=[
                col_func(start).expression,
                col_func(stop).expression,
                col_func(step).expression if step else expression.Literal.number(1),
            ],
        )
    )


def sequence_from_generate_array(
    start: ColumnOrName, stop: ColumnOrName, step: t.Optional[ColumnOrName] = None
) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="GENERATE_ARRAY",
            expressions=[
                col_func(start).expression,
                col_func(stop).expression,
                col_func(step).expression if step else expression.Literal.number(1),
            ],
        )
    )


def sequence_from_array_generate_range(
    start: ColumnOrName, stop: ColumnOrName, step: t.Optional[ColumnOrName] = None
) -> Column:
    col_func = get_func_from_session("col")
    when = get_func_from_session("when")
    lit = get_func_from_session("lit")

    return Column(
        expression.Anonymous(
            this="ARRAY_GENERATE_RANGE",
            expressions=[
                col_func(start).expression,
                (
                    col_func(stop) + when(col_func(stop) > lit(0), lit(1)).otherwise(lit(-1))
                ).expression,
                col_func(step).expression if step else lit(1).expression,
            ],
        )
    )


def regexp_extract_only_one_group(
    str: ColumnOrName, pattern: str, idx: t.Optional[int] = None
) -> Column:
    from sqlframe.base.functions import regexp_extract

    if re.compile(pattern).groups > 1 or (idx is not None and idx > 1):
        raise ValueError("This dialect only supports regular expressions with a single group")

    return regexp_extract(str, pattern, 1)


def hex_casted_as_bytes(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="TO_HEX",
            expressions=[col_func(col).cast("bytes", dialect="bigquery").expression],
        )
    )


def hex_using_encode(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="HEX_ENCODE",
            expressions=[col_func(col).expression],
        )
    )


def unhex_hex_decode_str(col: ColumnOrName) -> Column:
    col_func = get_func_from_session("col")

    return Column(
        expression.Anonymous(
            this="HEX_DECODE_STRING",
            expressions=[col_func(col).expression],
        )
    )


def bit_length_from_length(col: ColumnOrName) -> Column:
    lit = get_func_from_session("lit")
    col_func = get_func_from_session("col")

    return Column(expression.Length(this=col_func(col).expression)) * lit(8)


def any_value_always_ignore_nulls(
    col: ColumnOrName, ignoreNulls: t.Optional[t.Union[bool, Column]] = None
) -> Column:
    from sqlframe.base.functions import any_value

    if not ignoreNulls:
        logger.warning("Nulls are always ignored when using `ANY_VALUE` on this engine")
    return any_value(col)


def any_value_ignore_nulls_not_supported(
    col: ColumnOrName, ignoreNulls: t.Optional[t.Union[bool, Column]] = None
) -> Column:
    from sqlframe.base.functions import any_value

    if ignoreNulls:
        logger.warning("Ignoring nulls is not supported in this dialect")
    return any_value(col)


def current_user_from_session_user() -> Column:
    return Column(expression.Anonymous(this="SESSION_USER"))


def extract_convert_to_var(field: ColumnOrName, source: ColumnOrName) -> Column:
    from sqlframe.base.functions import extract

    field = expression.Var(this=Column.ensure_col(field).alias_or_name)  # type: ignore
    return extract(field, source)  # type: ignore


def left_cast_len(str: ColumnOrName, len: ColumnOrName) -> Column:
    from sqlframe.base.functions import left

    len = Column.ensure_col(len).cast("integer")
    return left(str, len)


def right_cast_len(str: ColumnOrName, len: ColumnOrName) -> Column:
    from sqlframe.base.functions import right

    len = Column.ensure_col(len).cast("integer")
    return right(str, len)


def position_cast_start(
    substr: ColumnOrName, str: ColumnOrName, start: t.Optional[ColumnOrName] = None
) -> Column:
    from sqlframe.base.functions import position

    start = Column.ensure_col(start).cast("integer") if start else None
    return position(substr, str, start)


def position_as_strpos(
    substr: ColumnOrName, str: ColumnOrName, start: t.Optional[ColumnOrName] = None
) -> Column:
    substr_func = get_func_from_session("substr")
    lit = get_func_from_session("lit")

    if start:
        str = substr_func(str, start)
    column = Column.invoke_anonymous_function(str, "STRPOS", substr)
    if start:
        return column + Column(start) - lit(1)
    return column


def to_number_using_to_double(col: ColumnOrName, format: ColumnOrName) -> Column:
    return Column.invoke_anonymous_function(col, "TO_DOUBLE", format)


def try_element_at_zero_based(col: ColumnOrName, extraction: ColumnOrName) -> Column:
    from sqlframe.base.functions import try_element_at

    lit = get_func_from_session("lit")
    index = Column.ensure_col(extraction)
    if isinstance(index.expression, expression.Literal) and index.expression.is_number:
        index = index - lit(1)
    return try_element_at(col, index)


def to_unix_timestamp_include_default_format(
    timestamp: ColumnOrName,
    format: t.Optional[ColumnOrName] = None,
) -> Column:
    from sqlframe.base.functions import to_unix_timestamp
    from sqlframe.base.session import _BaseSession

    return to_unix_timestamp(timestamp, format or _BaseSession().default_time_format)


def array_append_list_append(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    lit = get_func_from_session("lit")
    value = value if isinstance(value, Column) else lit(value)
    return Column.invoke_anonymous_function(col, "LIST_APPEND", value)


def array_append_using_array_cat(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    lit = get_func_from_session("lit")
    array = get_func_from_session("array")
    value = value if isinstance(value, Column) else lit(value)
    return Column.invoke_anonymous_function(col, "ARRAY_CONCAT", array(value))


def day_with_try_to_timestamp(col: ColumnOrName) -> Column:
    from sqlframe.base.functions import day

    try_to_timestamp = get_func_from_session("try_to_timestamp")
    to_date = get_func_from_session("to_date")
    when = get_func_from_session("when")
    _is_string = get_func_from_session("_is_string")
    coalesce = get_func_from_session("coalesce")
    return day(
        when(
            _is_string(col),
            coalesce(try_to_timestamp(col), to_date(col)),
        ).otherwise(col)
    )


def endswith_with_underscore(str: ColumnOrName, suffix: ColumnOrName) -> Column:
    return Column.invoke_anonymous_function(str, "ENDS_WITH", suffix)


def endswith_using_like(str: ColumnOrName, suffix: ColumnOrName) -> Column:
    concat = get_func_from_session("concat")
    lit = get_func_from_session("lit")

    return Column.invoke_expression_over_column(
        str, expression.Like, expression=concat(lit("%"), suffix)
    )


def try_to_timestamp_strptime(col: ColumnOrName, format: t.Optional[ColumnOrName] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    return Column.invoke_anonymous_function(col, "TRY_STRPTIME", _BaseSession().format_time(format))  # type: ignore


def try_to_timestamp_safe(col: ColumnOrName, format: t.Optional[ColumnOrName] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    return Column.invoke_anonymous_function(
        _BaseSession().format_time(format),  # type: ignore
        "SAFE.PARSE_TIMESTAMP",
        col,  # type: ignore
    )


def try_to_timestamp_pgtemp(col: ColumnOrName, format: t.Optional[ColumnOrName] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    return Column.invoke_anonymous_function(
        col,
        "pg_temp.TRY_TO_TIMESTAMP",
        _BaseSession().format_execution_time(format),  # type: ignore
    )


def typeof_pg_typeof(col: ColumnOrName) -> Column:
    return (
        Column.invoke_anonymous_function(col, "pg_typeof")
        .cast(expression.DataType(this=expression.DataType.Type.USERDEFINED, kind="regtype"))
        .cast("text")
    )


def typeof_from_variant(col: ColumnOrName) -> Column:
    col = Column.invoke_anonymous_function(col, "TO_VARIANT")
    return Column.invoke_anonymous_function(col, "TYPEOF")


def regexp_replace_global_option(
    str: ColumnOrName, pattern: str, replacement: str, position: t.Optional[int] = None
) -> Column:
    lit = get_func_from_session("lit")

    if position is not None:
        return Column.invoke_expression_over_column(
            str,
            expression.RegexpReplace,
            expression=lit(pattern),
            replacement=lit(replacement),
            position=lit(position),
            modifiers=lit("g"),
        )
    return Column.invoke_expression_over_column(
        str,
        expression.RegexpReplace,
        expression=lit(pattern),
        replacement=lit(replacement),
        modifiers=lit("g"),
    )


def _is_string_using_typeof_varchar(col: ColumnOrName) -> Column:
    typeof = get_func_from_session("typeof")
    lit = get_func_from_session("lit")
    return lit(typeof(col) == lit("VARCHAR"))


def _is_string_using_typeof_char_varying(col: ColumnOrName) -> Column:
    typeof = get_func_from_session("typeof")
    lit = get_func_from_session("lit")
    return lit(
        (typeof(col) == lit("text"))
        | (typeof(col) == lit("character varying"))
        | (typeof(col) == lit("unknown"))
        | (typeof(col) == lit("text"))
    )


def _is_string_using_typeof_string(col: ColumnOrName) -> Column:
    typeof = get_func_from_session("typeof")
    lit = get_func_from_session("lit")
    return lit(typeof(col) == lit("STRING"))


def _is_string_using_typeof_string_lcase(col: ColumnOrName) -> Column:
    typeof = get_func_from_session("typeof")
    lit = get_func_from_session("lit")
    return lit(typeof(col) == lit("string"))


def _is_integer_using_func(col: ColumnOrName) -> Column:
    return Column.invoke_anonymous_function(col, "IS_INTEGER")



from __future__ import annotations

import typing as t

from sqlframe.base.udf import _BaseUDFRegistration

if t.TYPE_CHECKING:
    from sqlframe.databricks.session import DatabricksSession


class DatabricksUDFRegistration(_BaseUDFRegistration["DatabricksSession"]): ...



import math
import typing as t

import pytest

from sqlframe.base.dataframe import _BaseDataFrame
from sqlframe.postgres import PostgresDataFrame
from sqlframe.snowflake import SnowflakeDataFrame

pytest_plugins = ["tests.integration.fixtures"]


def test_approx_quantile(get_engine_df_and_pyspark: t.Callable[[str], _BaseDataFrame]):
    employee = get_engine_df_and_pyspark("employee")
    if isinstance(employee, PostgresDataFrame):
        pytest.skip("Approx quantile is not supported by the engine: postgres")
    results = employee.stat.approxQuantile(["employee_id", "age"], [0.1, 0.5, 1.0], 0)
    if isinstance(employee, SnowflakeDataFrame):
        expected = [[1.4, 3.0, 5.0], [27.8, 37.0, 65.0]]
    else:
        expected = [[1.0, 3.0, 5.0], [27.0, 37.0, 65.0]]
    assert results == expected


def test_corr(get_engine_df_and_pyspark: t.Callable[[str], _BaseDataFrame]):
    employee = get_engine_df_and_pyspark("employee")
    results = employee.stat.corr("employee_id", "age")
    assert math.isclose(results, -0.5605569890127448)


def test_cov(get_engine_df_and_pyspark: t.Callable[[str], _BaseDataFrame]):
    employee = get_engine_df_and_pyspark("employee")
    results = employee.stat.cov("employee_id", "age")
    assert results == -13.5



import inspect
import sys

import sqlframe.base.functions

module = sys.modules["sqlframe.base.functions"]
globals().update(
    {
        name: func
        for name, func in inspect.getmembers(module, inspect.isfunction)
        if hasattr(func, "unsupported_engines")
        and "snowflake" not in func.unsupported_engines
        and "*" not in func.unsupported_engines
    }
)


from sqlframe.base.function_alternatives import (  # noqa
    any_value_ignore_nulls_not_supported as any_value,
    e_literal as e,
    expm1_from_exp as expm1,
    log1p_from_log as log1p,
    rint_from_round as rint,
    bitwise_not_from_bitnot as bitwise_not,
    skewness_from_skew as skewness,
    isnan_using_equal as isnan,
    isnull_using_equal as isnull,
    nanvl_as_case as nanvl,
    percentile_approx_without_accuracy_and_max_array as percentile_approx,
    bround_using_half_even as bround,
    shiftleft_from_bitshiftleft as shiftleft,
    shiftright_from_bitshiftright as shiftright,
    struct_with_eq as struct,
    make_date_date_from_parts as make_date,
    date_add_no_date_sub as date_add,
    date_add_no_date_sub as dateadd,
    date_sub_by_date_add as date_sub,
    add_months_using_func as add_months,
    months_between_cast_as_date_cast_roundoff as months_between,
    last_day_with_cast as last_day,
    from_unixtime_from_timestamp as from_unixtime,
    unix_timestamp_from_extract as unix_timestamp,
    base64_from_base64_encode as base64,
    unbase64_from_base64_decode_string as unbase64,
    format_number_from_to_char as format_number,
    overlay_from_substr as overlay,
    levenshtein_edit_distance as levenshtein,
    split_with_split as split,
    regexp_extract_coalesce_empty_str as regexp_extract,
    hex_using_encode as hex,
    unhex_hex_decode_str as unhex,
    create_map_with_cast as create_map,
    array_contains_cast_variant as array_contains,
    arrays_overlap_as_plural as arrays_overlap,
    slice_as_array_slice as slice,
    array_join_no_null_replacement as array_join,
    array_position_cast_variant_and_flip as array_position,
    element_at_using_brackets as element_at,
    array_intersect_using_intersection as array_intersect,
    array_union_using_array_concat as array_union,
    sort_array_using_array_sort as sort_array,
    flatten_using_array_flatten as flatten,
    map_concat_using_map_cat as map_concat,
    sequence_from_array_generate_range as sequence,
    to_number_using_to_double as to_number,
    typeof_from_variant as typeof,
    to_date_time_format as to_date,
    _is_integer_using_func as _is_integer,
)



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'spark' folder.

from __future__ import annotations

import datetime
import typing as t

from sqlglot import expressions as exp

if t.TYPE_CHECKING:
    from sqlframe.base.column import Column
    from sqlframe.base.types import DataType, StructType

PrimitiveType = t.Union[bool, float, int, str]
ColumnLiterals = t.Union[
    str, float, int, bool, t.List, t.Tuple, datetime.date, datetime.datetime, None
]
ColumnOrName = t.Union[Column, str]
ColumnOrLiteral = t.Union[
    Column, str, float, int, bool, t.List, t.Tuple, datetime.date, datetime.datetime, dict
]
SchemaInput = t.Union[str, t.List[str], t.Tuple[str, ...], StructType, t.Dict[str, t.Optional[str]]]
OutputExpressionContainer = t.Union[exp.Select, exp.Create, exp.Insert]
StorageLevel = str
PathOrPaths = t.Union[str, t.List[str]]
OptionalPrimitiveType = t.Optional[PrimitiveType]
DataTypeOrString = t.Union[DataType, str]


class UserDefinedFunctionLike(t.Protocol):
    func: t.Callable[..., t.Any]
    evalType: int
    deterministic: bool

    @property
    def returnType(self) -> DataType: ...

    def __call__(self, *args: ColumnOrName) -> Column: ...

    def asNondeterministic(self) -> UserDefinedFunctionLike: ...



import datetime

from sqlframe.standalone import functions as F
from sqlframe.standalone.window import Window


def test_simple():
    assert (F.col("cola") == 1).sql() == "cola = 1"


def test_neq():
    assert (F.col("cola") != 1).sql() == "cola <> 1"


def test_gt():
    assert (F.col("cola") > 1).sql() == "cola > 1"


def test_lt():
    assert (F.col("cola") < 1).sql() == "cola < 1"


def test_le():
    assert (F.col("cola") <= 1).sql() == "cola <= 1"


def test_ge():
    assert (F.col("cola") >= 1).sql() == "cola >= 1"


def test_and():
    assert (
        (F.col("cola") == F.col("colb")) & (F.col("colc") == F.col("cold"))
    ).sql() == "cola = colb AND colc = cold"


def test_or():
    assert (
        (F.col("cola") == F.col("colb")) | (F.col("colc") == F.col("cold"))
    ).sql() == "cola = colb OR colc = cold"


def test_mod():
    assert (F.col("cola") % 2).sql() == "(cola % 2)"


def test_add():
    assert (F.col("cola") + 1).sql() == "(cola + 1)"


def test_sub():
    assert (F.col("cola") - 1).sql() == "(cola - 1)"


def test_mul():
    assert (F.col("cola") * 2).sql() == "(cola * 2)"


def test_div():
    assert (F.col("cola") / 2).sql() == "(cola / 2)"


def test_radd():
    assert (1 + F.col("cola")).sql() == "(1 + cola)"


def test_rsub():
    assert (1 - F.col("cola")).sql() == "(1 - cola)"


def test_rmul():
    assert (1 * F.col("cola")).sql() == "(1 * cola)"


def test_rdiv():
    assert (1 / F.col("cola")).sql() == "(1 / cola)"


def test_pow():
    assert (F.col("cola") ** 2).sql() == "POWER(cola, 2)"


def test_rpow():
    assert (2 ** F.col("cola")).sql() == "POWER(2, cola)"


def test_invert():
    assert (~F.col("cola")).sql() == "NOT (cola)"


def test_invert_conjuction():
    assert (~(F.col("cola") | F.col("colb"))).sql() == "NOT (cola OR colb)"


def test_paren():
    assert ((F.col("cola") - F.col("colb")) / F.col("colc")).sql() == "((cola - colb) / colc)"


def test_implicit_paren():
    assert (F.col("cola") - F.col("colb") / F.col("colc")).sql() == "(cola - (colb / colc))"


def test_startswith():
    assert (F.col("cola").startswith("test")).sql() == "STARTSWITH(cola, 'test')"


def test_endswith():
    assert (F.col("cola").endswith("test")).sql() == "ENDSWITH(cola, 'test')"


def test_rlike():
    assert (F.col("cola").rlike("foo")).sql() == "cola RLIKE 'foo'"


def test_like():
    assert (F.col("cola").like("foo%")).sql() == "cola LIKE 'foo%'"


def test_ilike():
    assert (F.col("cola").ilike("foo%")).sql() == "cola ILIKE 'foo%'"


def test_substring():
    assert (F.col("cola").substr(2, 3)).sql() == "SUBSTRING(cola, 2, 3)"


def test_isin():
    assert (F.col("cola").isin([1, 2, 3])).sql() == "cola IN (1, 2, 3)"
    assert (F.col("cola").isin(1, 2, 3)).sql() == "cola IN (1, 2, 3)"


def test_asc():
    assert (F.col("cola").asc()).sql() == "cola ASC"


def test_desc():
    assert (F.col("cola").desc()).sql() == "cola DESC"


def test_asc_nulls_first():
    assert (F.col("cola").asc_nulls_first()).sql() == "cola ASC"


def test_asc_nulls_last():
    assert (F.col("cola").asc_nulls_last()).sql() == "cola ASC NULLS LAST"


def test_desc_nulls_first():
    assert (F.col("cola").desc_nulls_first()).sql() == "cola DESC NULLS FIRST"


def test_desc_nulls_last():
    assert (F.col("cola").desc_nulls_last()).sql() == "cola DESC"


def test_when_otherwise():
    assert (F.when(F.col("cola") == 1, 2)).sql() == "CASE WHEN cola = 1 THEN 2 END"
    assert (F.col("cola").when(F.col("cola") == 1, 2)).sql() == "CASE WHEN cola = 1 THEN 2 END"
    assert (
        F.when(F.col("cola") == 1, 2).when(F.col("colb") == 2, 3)
    ).sql() == "CASE WHEN cola = 1 THEN 2 WHEN colb = 2 THEN 3 END"
    assert (
        F.col("cola").when(F.col("cola") == 1, 2).when(F.col("colb") == 2, 3).sql()
        == "CASE WHEN cola = 1 THEN 2 WHEN colb = 2 THEN 3 END"
    )
    assert (
        F.when(F.col("cola") == 1, 2).when(F.col("colb") == 2, 3).otherwise(4).sql()
        == "CASE WHEN cola = 1 THEN 2 WHEN colb = 2 THEN 3 ELSE 4 END"
    )


def test_is_null():
    assert (F.col("cola").isNull()).sql() == "cola IS NULL"


def test_is_not_null():
    assert (F.col("cola").isNotNull()).sql() == "NOT cola IS NULL"


def test_cast():
    assert (F.col("cola").cast("INT")).sql() == "CAST(cola AS INT)"


def test_alias():
    assert (F.col("cola").alias("new_name")).sql() == "cola AS new_name"


def test_between():
    assert (F.col("cola").between(1, 3)).sql() == "cola BETWEEN 1 AND 3"
    assert (F.col("cola").between(10.1, 12.1)).sql() == "cola BETWEEN 10.1 AND 12.1"
    assert (
        F.col("cola").between(datetime.date(2022, 1, 1), datetime.date(2022, 3, 1))
    ).sql() == "cola BETWEEN CAST('2022-01-01' AS DATE) AND CAST('2022-03-01' AS DATE)"
    assert (
        (
            F.col("cola").between(
                datetime.datetime(2022, 1, 1, 1, 1, 1), datetime.datetime(2022, 3, 1, 1, 1, 1)
            )
        ).sql()
        == "cola BETWEEN CAST('2022-01-01 01:01:01' AS TIMESTAMP) AND CAST('2022-03-01 01:01:01' AS TIMESTAMP)"
    )


def test_over():
    assert (
        (
            F.sum("cola").over(
                Window.partitionBy("colb").orderBy("colc").rowsBetween(1, Window.unboundedFollowing)
            )
        ).sql()
        == "SUM(cola) OVER (PARTITION BY colb ORDER BY colc ROWS BETWEEN 1 PRECEDING AND UNBOUNDED FOLLOWING)"
    )
    assert (
        (
            F.sum("cola").over(
                Window.partitionBy("colb")
                .orderBy("colc")
                .rangeBetween(1, Window.unboundedFollowing)
            )
        ).sql()
        == "SUM(cola) OVER (PARTITION BY colb ORDER BY colc RANGE BETWEEN 1 PRECEDING AND UNBOUNDED FOLLOWING)"
    )


def test_get_item():
    assert F.col("cola").getItem(1).sql() == "ELEMENT_AT(cola, (1 + 1))"



from sqlframe.databricks.catalog import DatabricksCatalog
from sqlframe.databricks.column import Column
from sqlframe.databricks.dataframe import (
    DatabricksDataFrame,
    DatabricksDataFrameNaFunctions,
    DatabricksDataFrameStatFunctions,
)
from sqlframe.databricks.group import DatabricksGroupedData
from sqlframe.databricks.readwriter import (
    DatabricksDataFrameReader,
    DatabricksDataFrameWriter,
)
from sqlframe.databricks.session import DatabricksSession
from sqlframe.databricks.types import Row
from sqlframe.databricks.udf import DatabricksUDFRegistration
from sqlframe.databricks.window import Window, WindowSpec

__all__ = [
    "Column",
    "Row",
    "DatabricksCatalog",
    "DatabricksDataFrame",
    "DatabricksDataFrameNaFunctions",
    "DatabricksGroupedData",
    "DatabricksDataFrameReader",
    "DatabricksDataFrameWriter",
    "DatabricksSession",
    "DatabricksDataFrameStatFunctions",
    "DatabricksUDFRegistration",
    "Window",
    "WindowSpec",
]



import pytest

from sqlframe.standalone import types


@pytest.mark.parametrize(
    "dtype, expected",
    [
        (types.StringType(), "string"),
        (types.CharType(100), "char"),
        (types.VarcharType(65), "varchar"),
        (types.BinaryType(), "binary"),
        (types.BooleanType(), "boolean"),
        (types.DateType(), "date"),
        (types.TimestampType(), "timestamp"),
        (types.TimestampNTZType(), "timestamp_ntz"),
        (types.DecimalType(10, 3), "decimal(10, 3)"),
        (types.DoubleType(), "double"),
        (types.FloatType(), "float"),
        (types.ByteType(), "byte"),
        (types.IntegerType(), "integer"),
        (types.LongType(), "long"),
        (types.ShortType(), "short"),
        (types.ArrayType(types.IntegerType()), "array<integer>"),
        (types.MapType(types.IntegerType(), types.StringType()), "map<integer, string>"),
        (
            types.StructType(
                [
                    types.StructField("cola", types.IntegerType()),
                    types.StructField("colb", types.StringType()),
                ]
            ),
            "struct<cola:integer, colb:string>",
        ),
    ],
)
def test_string(dtype, expected):
    assert dtype.simpleString() == expected



from __future__ import annotations

import typing as t

from sqlframe.base.udf import _BaseUDFRegistration

if t.TYPE_CHECKING:
    from sqlframe.base._typing import DataTypeOrString, UserDefinedFunctionLike
    from sqlframe.spark.session import SparkSession  # type: ignore


class SparkUDFRegistration(_BaseUDFRegistration["SparkSession"]):
    def register(
        self,
        name: str,
        f: t.Union[t.Callable[..., t.Any], UserDefinedFunctionLike],
        returnType: t.Optional[DataTypeOrString] = None,
    ) -> UserDefinedFunctionLike:
        return self.sparkSession.spark_session.udf.register(name, f, returnType=returnType)  # type: ignore

    def registerJavaFunction(
        self,
        name: str,
        javaClassName: str,
        returnType: t.Optional[DataTypeOrString] = None,
    ) -> None:
        self.sparkSession.spark_session.udf.registerJavaFunction(
            name,
            javaClassName,
            returnType=returnType,  # type: ignore
        )

    def registerJavaUDAF(self, name: str, javaClassName: str) -> None:
        self.sparkSession.spark_session.udf.registerJavaUDAF(name, javaClassName)



from sqlframe.duckdb.catalog import DuckDBCatalog
from sqlframe.duckdb.column import Column
from sqlframe.duckdb.dataframe import (
    DuckDBDataFrame,
    DuckDBDataFrameNaFunctions,
    DuckDBDataFrameStatFunctions,
)
from sqlframe.duckdb.group import DuckDBGroupedData
from sqlframe.duckdb.readwriter import DuckDBDataFrameReader, DuckDBDataFrameWriter
from sqlframe.duckdb.session import DuckDBSession
from sqlframe.duckdb.types import Row
from sqlframe.duckdb.udf import DuckDBUDFRegistration
from sqlframe.duckdb.window import Window, WindowSpec

__all__ = [
    "Column",
    "DuckDBCatalog",
    "DuckDBDataFrame",
    "DuckDBDataFrameNaFunctions",
    "DuckDBGroupedData",
    "DuckDBDataFrameReader",
    "DuckDBDataFrameWriter",
    "DuckDBSession",
    "DuckDBDataFrameStatFunctions",
    "DuckDBUDFRegistration",
    "Row",
    "Window",
    "WindowSpec",
]



from __future__ import annotations

import os
import typing as t

import pytest
from sqlglot import exp

from sqlframe.snowflake.session import SnowflakeSession

if t.TYPE_CHECKING:
    from snowflake.connector import SnowflakeConnection

pytest_plugins = ["tests.common_fixtures"]
pytestmark = [
    pytest.mark.snowflake,
    pytest.mark.xdist_group("snowflake_tests"),
]


@pytest.fixture
def cleanup_connector() -> t.Iterator[SnowflakeConnection]:
    from snowflake.connector import connect

    conn = connect(
        account=os.environ["SQLFRAME_SNOWFLAKE_ACCOUNT"],
        user=os.environ["SQLFRAME_SNOWFLAKE_USER"],
        password=os.environ["SQLFRAME_SNOWFLAKE_PASSWORD"],
        warehouse=os.environ["SQLFRAME_SNOWFLAKE_WAREHOUSE"],
        database=os.environ["SQLFRAME_SNOWFLAKE_DATABASE"],
        schema=os.environ["SQLFRAME_SNOWFLAKE_SCHEMA"],
    )
    conn.cursor().execute("CREATE SCHEMA IF NOT EXISTS db1")
    conn.cursor().execute("DROP TABLE IF EXISTS db1.test_table")
    conn.cursor().execute("CREATE TABLE IF NOT EXISTS db1.test_table (cola INT, colb STRING)")

    yield conn
    conn.cursor().execute("DROP TABLE IF EXISTS db1.test_table")


def test_session_from_config(cleanup_connector: SnowflakeConnection):
    session = SnowflakeSession.builder.config("sqlframe.conn", cleanup_connector).getOrCreate()
    columns = session.catalog.get_columns("sqlframe.db1.test_table")
    assert columns == {
        "cola": exp.DataType.build("DECIMAL(38, 0)", dialect=session.output_dialect),
        "colb": exp.DataType.build("TEXT", dialect=session.output_dialect),
    }



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t

from sqlframe.base.readerwriter import _BaseDataFrameReader, _BaseDataFrameWriter

if t.TYPE_CHECKING:
    from sqlframe.standalone.dataframe import StandaloneDataFrame
    from sqlframe.standalone.session import StandaloneSession


class StandaloneDataFrameReader(_BaseDataFrameReader["StandaloneSession", "StandaloneDataFrame"]):
    pass


class StandaloneDataFrameWriter(_BaseDataFrameWriter["StandaloneSession", "StandaloneDataFrame"]):
    pass



from sqlglot import exp

from sqlframe.duckdb.session import DuckDBSession


def test_session_from_config():
    import duckdb

    conn = duckdb.connect()
    conn.execute("CREATE TABLE test_table (cola INT, colb STRING)")
    session = DuckDBSession.builder.config("sqlframe.conn", conn).getOrCreate()
    columns = session.catalog.get_columns("test_table")
    assert columns == {"cola": exp.DataType.build("INT"), "colb": exp.DataType.build("TEXT")}



import pytest
from sqlglot import exp

from sqlframe.bigquery.session import BigQuerySession

pytestmark = [
    pytest.mark.bigquery,
    pytest.mark.xdist_group("bigquery_tests"),
]


def test_session_from_config():
    from google.cloud.bigquery.dbapi import connect

    conn = connect()
    conn.cursor().execute("CREATE SCHEMA IF NOT EXISTS db1")
    conn.cursor().execute("CREATE TABLE IF NOT EXISTS db1.test_table (cola INT, colb STRING)")
    session = BigQuerySession.builder.config("default_dataset", "sqlframe.db1").getOrCreate()
    columns = session.catalog.get_columns("db1.test_table")
    assert columns == {"cola": exp.DataType.build("BIGINT"), "colb": exp.DataType.build("TEXT")}



import pytest

from sqlframe.redshift import Column as RedshiftColumn
from sqlframe.redshift import (
    RedshiftCatalog,
    RedshiftDataFrame,
    RedshiftDataFrameNaFunctions,
    RedshiftDataFrameReader,
    RedshiftDataFrameStatFunctions,
    RedshiftDataFrameWriter,
    RedshiftGroupedData,
    RedshiftSession,
    RedshiftUDFRegistration,
)
from sqlframe.redshift import Row as RedshiftRow
from sqlframe.redshift import Window as RedshiftWindow
from sqlframe.redshift import WindowSpec as RedshiftWindowSpec
from sqlframe.redshift import functions as RedshiftF
from sqlframe.redshift import types as RedshiftTypes


@pytest.mark.forked
def test_activate_redshift(check_pyspark_imports):
    check_pyspark_imports(
        "redshift",
        sqlf_session=RedshiftSession,
        sqlf_catalog=RedshiftCatalog,
        sqlf_column=RedshiftColumn,
        sqlf_dataframe=RedshiftDataFrame,
        sqlf_grouped_data=RedshiftGroupedData,
        sqlf_window=RedshiftWindow,
        sqlf_window_spec=RedshiftWindowSpec,
        sqlf_functions=RedshiftF,
        sqlf_types=RedshiftTypes,
        sqlf_udf_registration=RedshiftUDFRegistration,
        sqlf_dataframe_reader=RedshiftDataFrameReader,
        sqlf_dataframe_writer=RedshiftDataFrameWriter,
        sqlf_dataframe_na_functions=RedshiftDataFrameNaFunctions,
        sqlf_dataframe_stat_functions=RedshiftDataFrameStatFunctions,
        sqlf_row=RedshiftRow,
    )



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import datetime
import math
import typing as t

import sqlglot
from sqlglot import Dialect
from sqlglot import expressions as exp
from sqlglot.helper import flatten, is_iterable
from sqlglot.optimizer.normalize_identifiers import normalize_identifiers

from sqlframe.base.exceptions import UnsupportedOperationError
from sqlframe.base.types import DataType
from sqlframe.base.util import get_func_from_session, quote_preserving_alias_or_name

if t.TYPE_CHECKING:
    from sqlframe.base._typing import ColumnOrLiteral, ColumnOrName
    from sqlframe.base.window import WindowSpec


class Column:
    def __init__(self, expression: t.Optional[t.Union[ColumnOrLiteral, exp.Expression]]):
        from sqlframe.base.session import _BaseSession

        if isinstance(expression, Column):
            expression = expression.expression  # type: ignore
        elif expression is None or not isinstance(expression, (str, exp.Expression)):
            expression = self._lit(expression).expression  # type: ignore
        elif not isinstance(expression, exp.Column):
            expression = sqlglot.maybe_parse(
                expression, dialect=_BaseSession().input_dialect
            ).transform(_BaseSession().input_dialect.normalize_identifier, copy=False)
        if expression is None:
            raise ValueError(f"Could not parse {expression}")

        self.expression: exp.Expression = expression  # type: ignore

    def __repr__(self):
        return repr(self.expression)

    def __hash__(self):
        return hash(self.expression)

    def __eq__(self, other: ColumnOrLiteral) -> Column:  # type: ignore
        return self.binary_op(exp.EQ, other)

    def __ne__(self, other: ColumnOrLiteral) -> Column:  # type: ignore
        return self.binary_op(exp.NEQ, other)

    def __gt__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.GT, other)

    def __ge__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.GTE, other)

    def __lt__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.LT, other)

    def __le__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.LTE, other)

    def __and__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.And, other)

    def __or__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.Or, other)

    def __mod__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.Mod, other, paren=True)

    def __add__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.Add, other, paren=True)

    def __sub__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.Sub, other, paren=True)

    def __mul__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.Mul, other, paren=True)

    def __truediv__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.Div, other, paren=True)

    def __div__(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.Div, other, paren=True)

    def __neg__(self) -> Column:
        return self.unary_op(exp.Neg)

    def __radd__(self, other: ColumnOrLiteral) -> Column:
        return self.inverse_binary_op(exp.Add, other, paren=True)

    def __rsub__(self, other: ColumnOrLiteral) -> Column:
        return self.inverse_binary_op(exp.Sub, other, paren=True)

    def __rmul__(self, other: ColumnOrLiteral) -> Column:
        return self.inverse_binary_op(exp.Mul, other, paren=True)

    def __rdiv__(self, other: ColumnOrLiteral) -> Column:
        return self.inverse_binary_op(exp.Div, other, paren=True)

    def __rtruediv__(self, other: ColumnOrLiteral) -> Column:
        return self.inverse_binary_op(exp.Div, other, paren=True)

    def __rmod__(self, other: ColumnOrLiteral) -> Column:
        return self.inverse_binary_op(exp.Mod, other, paren=True)

    def __pow__(self, power: ColumnOrLiteral, modulo=None):
        return Column(exp.Pow(this=self.expression, expression=Column(power).expression))

    def __rpow__(self, power: ColumnOrLiteral):
        return Column(exp.Pow(this=Column(power).expression, expression=self.expression))

    def __invert__(self):
        return self.unary_op(exp.Not)

    def __rand__(self, other: ColumnOrLiteral) -> Column:
        return self.inverse_binary_op(exp.And, other)

    def __ror__(self, other: ColumnOrLiteral) -> Column:
        return self.inverse_binary_op(exp.Or, other)

    def __call__(self, *args, **kwargs):
        raise UnsupportedOperationError(
            "Tried to call a column which is unexpected. Did you mean to call a method on a DataFrame? If so, make sure the method is typed correctly and is supported. If not, please open an issue requesting support: https://github.com/eakmanrq/sqlframe/issues"
        )

    @classmethod
    def ensure_col(cls, value: t.Optional[t.Union[ColumnOrName, exp.Expression]]) -> Column:
        col = get_func_from_session("col")

        return col(value)

    @classmethod
    def ensure_cols(cls, args: t.List[t.Union[ColumnOrName, exp.Expression]]) -> t.List[Column]:
        return [cls.ensure_col(x) if not isinstance(x, Column) else x for x in args]

    @classmethod
    def _lit(cls, value: ColumnOrLiteral) -> Column:
        from sqlframe.base.session import _BaseSession
        from sqlframe.base.types import Row

        if isinstance(value, Row):
            columns = [
                exp.PropertyEQ(
                    this=exp.to_identifier(k).transform(
                        _BaseSession().input_dialect.normalize_identifier, copy=False
                    ),
                    expression=cls._lit(v).expression,
                )
                for k, v in value.asDict().items()
            ]
            return cls(exp.Struct(expressions=columns))
        if isinstance(value, (list, set)):
            return cls(exp.Array(expressions=[cls._lit(x).expression for x in value]))
        if isinstance(value, tuple):
            return cls(exp.Tuple(expressions=[cls._lit(x).expression for x in value]))
        if isinstance(value, dict):
            return cls(
                exp.VarMap(
                    keys=exp.Array(expressions=[cls._lit(k).expression for k in value.keys()]),
                    values=exp.Array(expressions=[cls._lit(v).expression for v in value.values()]),
                )
            )
        elif value is not None and isinstance(value, float) and math.isnan(value):
            return cls(exp.cast(exp.Literal.string("NaN"), exp.DataType.build("float")))
        elif isinstance(value, datetime.datetime):
            if value.tzinfo is None:
                value = value.isoformat(sep=" ")
                return cls(exp.cast(exp.Literal.string(value), exp.DataType.Type.TIMESTAMP))
            else:
                value = value.astimezone(datetime.timezone.utc).isoformat(sep=" ")
                return cls(exp.cast(exp.Literal.string(value), exp.DataType.Type.TIMESTAMPTZ))
        return cls(exp.convert(value))

    @classmethod
    def invoke_anonymous_function(
        cls, column: t.Optional[ColumnOrName], func_name: str, *args: t.Optional[ColumnOrName]
    ) -> Column:
        columns = [] if column is None else [cls.ensure_col(column)]
        column_args = [cls.ensure_col(arg) for arg in args]
        expressions = [x.expression for x in columns + column_args]
        new_expression = exp.Anonymous(this=func_name.upper(), expressions=expressions)
        return Column(new_expression)

    @classmethod
    def invoke_expression_over_column(
        cls, column: t.Optional[ColumnOrName], callable_expression: t.Callable, **kwargs
    ) -> Column:
        ensured_column = None if column is None else cls.ensure_col(column)
        ensure_expression_values = {
            k: (
                [cls.ensure_col(x).expression for x in v]
                if is_iterable(v)
                else cls.ensure_col(v).expression
            )
            for k, v in kwargs.items()
            if v is not None
        }
        new_expression = (
            callable_expression(**ensure_expression_values)
            if ensured_column is None
            else callable_expression(
                this=ensured_column.column_expression, **ensure_expression_values
            )
        )
        return Column(new_expression)

    def binary_op(
        self, klass: t.Callable, other: ColumnOrLiteral, paren: bool = False, **kwargs
    ) -> Column:
        other = self._lit(other) if isinstance(other, str) else Column(other)
        op = klass(this=self.column_expression, expression=other.column_expression, **kwargs)
        if paren:
            return Column(exp.Paren(this=op))
        return Column(op)

    def inverse_binary_op(
        self, klass: t.Callable, other: ColumnOrLiteral, paren: bool = False, **kwargs
    ) -> Column:
        other = self._lit(other) if isinstance(other, str) else Column(other)
        op = klass(this=other.column_expression, expression=self.column_expression, **kwargs)
        if paren:
            return Column(exp.Paren(this=op))
        return Column(op)

    def unary_op(self, klass: t.Callable, **kwargs) -> Column:
        return Column(klass(this=exp.Paren(this=self.column_expression), **kwargs))

    @property
    def is_alias(self):
        return isinstance(self.expression, exp.Alias)

    @property
    def is_column(self):
        return isinstance(self.expression, exp.Column)

    @property
    def column_expression(self) -> t.Union[exp.Column, exp.Literal]:
        return self.expression.unalias()

    @property
    def alias_or_name(self) -> str:
        return quote_preserving_alias_or_name(self.expression)  # type: ignore

    @property
    def column_alias_or_name(self) -> str:
        return quote_preserving_alias_or_name(self.column_expression)  # type: ignore

    @classmethod
    def ensure_literal(cls, value) -> Column:
        from sqlframe.base.functions import lit

        if isinstance(value, cls):
            value = value.expression
        if not isinstance(value, exp.Literal):
            return lit(value)
        return Column(value)

    @property
    def dtype(self) -> t.Optional[DataType]:
        expression = self.expression.unalias()
        if isinstance(expression, exp.Cast):
            return expression.args.get("to")
        return None

    def copy(self) -> Column:
        return Column(self.expression.copy())

    def set_table_name(self, table_name: str, copy=False) -> Column:
        expression = self.expression.copy() if copy else self.expression
        expression.set("table", exp.to_identifier(table_name))
        return Column(expression)

    def sql(self, **kwargs) -> str:
        from sqlframe.base.session import _BaseSession

        return self.expression.sql(**{"dialect": _BaseSession().input_dialect, **kwargs})

    def alias(self, name: str) -> Column:
        from sqlframe.base.session import _BaseSession

        dialect = _BaseSession().input_dialect
        alias: exp.Expression = normalize_identifiers(
            exp.parse_identifier(name, dialect=dialect), dialect=dialect
        )
        new_expression = exp.Alias(
            this=self.column_expression,
            alias=alias.this if isinstance(alias, exp.Column) else alias,
        )
        return Column(new_expression)

    def asc(self) -> Column:
        new_expression = exp.Ordered(this=self.column_expression, desc=False, nulls_first=True)
        return Column(new_expression)

    def desc(self) -> Column:
        new_expression = exp.Ordered(this=self.column_expression, desc=True, nulls_first=False)
        return Column(new_expression)

    asc_nulls_first = asc

    def asc_nulls_last(self) -> Column:
        new_expression = exp.Ordered(this=self.column_expression, desc=False, nulls_first=False)
        return Column(new_expression)

    def desc_nulls_first(self) -> Column:
        new_expression = exp.Ordered(this=self.column_expression, desc=True, nulls_first=True)
        return Column(new_expression)

    desc_nulls_last = desc

    def when(self, condition: Column, value: t.Any) -> Column:
        from sqlframe.base.functions import when

        column_with_if = when(condition, value)
        if not isinstance(self.expression, exp.Case):
            return column_with_if
        new_column = self.copy()
        new_column.expression.args["ifs"].extend(column_with_if.expression.args["ifs"])
        return new_column

    def otherwise(self, value: t.Any) -> Column:
        from sqlframe.base.functions import lit

        true_value = value if isinstance(value, Column) else lit(value)
        new_column = self.copy()
        new_column.expression.set("default", true_value.column_expression)
        return new_column

    def isNull(self) -> Column:
        new_expression = exp.Is(this=self.column_expression, expression=exp.Null())
        return Column(new_expression)

    def isNotNull(self) -> Column:
        new_expression = exp.Not(this=exp.Is(this=self.column_expression, expression=exp.Null()))
        return Column(new_expression)

    def eqNullSafe(self, other: ColumnOrLiteral) -> Column:
        return self.binary_op(exp.NullSafeEQ, other)

    def cast(
        self,
        dataType: t.Union[str, DataType, exp.DataType, exp.DataType.Type],
        dialect: t.Optional[t.Union[str, Dialect]] = None,
    ) -> Column:
        from sqlframe.base.session import _BaseSession

        if isinstance(dataType, DataType):
            dataType = dataType.simpleString()
        return Column(
            exp.cast(
                self.column_expression, dataType, dialect=dialect or _BaseSession().input_dialect
            )
        )

    def startswith(self, value: t.Union[str, Column]) -> Column:
        value = self._lit(value) if not isinstance(value, Column) else value
        return self.invoke_expression_over_column(self, exp.StartsWith, expression=value.expression)

    def endswith(self, value: t.Union[str, Column]) -> Column:
        value = self._lit(value) if not isinstance(value, Column) else value
        return self.invoke_anonymous_function(self, "ENDSWITH", value)

    def rlike(self, regexp: str) -> Column:
        return self.invoke_expression_over_column(
            column=self, callable_expression=exp.RegexpLike, expression=self._lit(regexp).expression
        )

    def like(self, other: str):
        return self.invoke_expression_over_column(
            self, exp.Like, expression=self._lit(other).expression
        )

    def ilike(self, other: str):
        return self.invoke_expression_over_column(
            self, exp.ILike, expression=self._lit(other).expression
        )

    def substr(self, startPos: t.Union[int, Column], length: t.Union[int, Column]) -> Column:
        startPos = self._lit(startPos) if not isinstance(startPos, Column) else startPos
        length = self._lit(length) if not isinstance(length, Column) else length
        return Column.invoke_expression_over_column(
            self, exp.Substring, start=startPos.expression, length=length.expression
        )

    def isin(self, *cols: t.Union[ColumnOrLiteral, t.Iterable[ColumnOrLiteral]]):
        columns = flatten(cols) if isinstance(cols[0], (list, set, tuple)) else cols  # type: ignore
        expressions = [self._lit(x).expression for x in columns]  # type: ignore
        return Column.invoke_expression_over_column(self, exp.In, expressions=expressions)  # type: ignore

    def between(
        self,
        lowerBound: t.Union[ColumnOrLiteral],
        upperBound: t.Union[ColumnOrLiteral],
    ) -> Column:
        lower_bound_exp = (
            self._lit(lowerBound) if not isinstance(lowerBound, Column) else lowerBound
        )
        upper_bound_exp = (
            self._lit(upperBound) if not isinstance(upperBound, Column) else upperBound
        )
        return Column(
            exp.Between(
                this=self.column_expression,
                low=lower_bound_exp.expression,
                high=upper_bound_exp.expression,
            )
        )

    def over(self, window: WindowSpec) -> Column:
        window_expression = window.expression.copy()
        window_expression.set("this", self.column_expression)
        return Column(window_expression)

    def getItem(self, key: t.Any) -> Column:
        """
        An expression that gets an item at position ``ordinal`` out of a list,
        or gets an item by key out of a dict.

        .. versionadded:: 1.3.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        key
            a literal value, or a :class:`Column` expression.
            The result will only be true at a location if the item matches in the column.

             .. deprecated:: 3.0.0
                 :class:`Column` as a parameter is deprecated.

        Returns
        -------
        :class:`Column`
            Column representing the item(s) got at position out of a list or by key out of a dict.

        Examples
        --------
        >>> df = spark.createDataFrame([([1, 2], {"key": "value"})], ["l", "d"])
        >>> df.select(df.l.getItem(0), df.d.getItem("key")).show()
        +----+------+
        |l[0]|d[key]|
        +----+------+
        |   1| value|
        +----+------+
        """
        element_at = get_func_from_session("element_at")
        lit = get_func_from_session("lit")
        key = lit(key) if not isinstance(key, Column) else key
        if isinstance(key.expression, exp.Literal) and key.expression.is_number:
            key = key + lit(1)
        return element_at(self, key)
